# RCAIDE/Library/Methods/Powertrain/Converters/Turbine/__init__.py

"""
This module provides functionality for modeling turbines in powertrains. It includes methods for computing 
turbine performance and appending turbine conditions to simulation results.

See Also
--------
RCAIDE.Library.Methods.Powertrain.Converters
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .append_turbine_conditions   import append_turbine_conditions                     
from .compute_turbine_performance import compute_turbine_performance


# RCAIDE/Methods/Mission/Common/Residuals/__init__.py
# 

"""
This module contains (a) residual functions used in mission analysis for RCAIDE. It provides 
flight dynamics residual calculations that are essential for solving flight dynamics 
equations during mission simulation.

See Also
--------
RCAIDE.Library.Mission.Segments : Mission segment definitions that may use these residuals
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# -----------------------------------------------------------------------------------------------------------------------Common
   
from .flight_dynamics import flight_dynamics  


# RCAIDE/Library/Methods/Powertrain/Converters/Combustor/append_combustor_conditions.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke  

from RCAIDE.Framework.Mission.Common     import   Conditions

# ---------------------------------------------------------------------------------------------------------------------- 
#  append_combustor_conditions
# ----------------------------------------------------------------------------------------------------------------------    
def append_combustor_conditions(combustor, segment, energy_conditions):
    """
    Initializes combustor operating conditions for a mission segment.
    
    Parameters
    ----------
    combustor : RCAIDE.Library.Components.Converters.Combustor
        Combustor component with the following attributes:
            - tag : str
                Identifier for the combustor
    segment : RCAIDE.Framework.Mission.Segments.Segment
        Mission segment with the following attributes:
            - state : Data
                Segment state
                    - ones_row : function
                        Function to create array of ones with specified length
    energy_conditions : RCAIDE.Framework.Mission.Common.Conditions
        Energy conditions container where combustor conditions will be stored
    
    Returns
    -------
    None
        Results are stored in energy_conditions.converters[combustor.tag]
    
    Notes
    -----
    This function initializes the necessary data structures for storing combustor
    operating conditions during a mission segment. It creates a container for the
    combustor in the energy conditions and initializes the non-dimensional mass ratio
    with ones.
    
    The function initializes the following in energy_conditions.converters[combustor.tag]:
        - inputs : Conditions
            Input conditions container
                - nondim_mass_ratio : numpy.ndarray
                    Non-dimensional mass ratio, initialized with ones
        - outputs : Conditions
            Output conditions container (empty)
    
    The non-dimensional mass ratio represents the ratio of mass flow at the combustor exit
    to the mass flow at the combustor inlet, accounting for the addition of fuel.
    
    **Major Assumptions**
        * Non-dimensional mass ratio is initialized with ones (no fuel addition initially)
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Combustor.compute_combustor_performance
    """
    ones_row    = segment.state.ones_row 
    energy_conditions.converters[combustor.tag]                           = Conditions() 
    energy_conditions.converters[combustor.tag].inputs                    = Conditions() 
    energy_conditions.converters[combustor.tag].inputs.nondim_mass_ratio  = ones_row(1)
    energy_conditions.converters[combustor.tag].outputs                   = Conditions()
    return 


# RCAIDE/Library/Methods/Powertrain/Converters/Ducted_Fan/compute_ducted_fan_performance.py

# 
# Created:  Jan 2025, M. Clarke
# Modified: Jan 2025, M. Clarke, M. Guidotti    

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports
from RCAIDE.Framework.Core    import Data , Units, orientation_product, orientation_transpose  

# package imports
import  numpy as  np 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Generalized Rotor Class
# ---------------------------------------------------------------------------------------------------------------------- 
def BEMT_performance(ducted_fan,conditions):

    """
    Computes ducted fan performance characteristics Blade Element Momentum Theory (BEMT)  

    Parameters
    ----------
    propulsor : Converter
        Ducted fan propulsor component containing the ducted fan
    state : Conditions
        Mission segment state conditions

    Returns
    -------
    None
        Updates state.conditions.energy.converters[ducted_fan.tag] with computed performance data:
            - thrust : array(N,3)
                Thrust vector [N]
            - power : array(N,1)
                Power required [W]
            - torque : array(N,1)
                Shaft torque [N-m]
            - moment : array(N,3)
                Moment vector [N-m]
            - efficiency : array(N,1)
                Propulsive efficiency [-]
            - tip_mach : array(N,1)
                Blade tip Mach number [-]
            - thrust_coefficient : array(N,1)
                Non-dimensional thrust coefficient [-]
            - power_coefficient : array(N,1)
                Non-dimensional power coefficient [-]
            - figure_of_merit : array(N,1)
                Hovering figure of merit [-] (BEMT only)

    Notes
    -----  
    **Major Assumptions**
        * Steady state operation
        * Incompressible flow for Rankine-Froude theory
        * Rigid blades
        * No blade-wake interaction
        * No ground effect

    See Also
    --------
    RCAIDE.Library.Components.Powertrain.Converters.Ducted_Fan
    RCAIDE.Library.Methods.Powertrain.Converters.Ducted_Fan.design_ducted_fan
    """
  

    a                     = conditions.freestream.speed_of_sound 
    rho                   = conditions.freestream.density  
    commanded_TV          = conditions.energy.converters[ducted_fan.tag].commanded_thrust_vector_angle
    omega                 = conditions.energy.converters[ducted_fan.tag].omega   
    alt                   = conditions.freestream.altitude 
     
    altitude       = alt/ 1000  
    n              = omega/(2.*np.pi)   # Rotations per second
    D              = ducted_fan.tip_radius * 2
    A              = 0.25 * np.pi * (D ** 2)
    
    # Unpack freestream conditions 
    Vv             = conditions.frames.inertial.velocity_vector 

    # Number of radial stations and segment control point
    B              = ducted_fan.number_of_rotor_blades
    Nr             = ducted_fan.number_of_radial_stations
    ctrl_pts       = len(Vv)
     
    # Velocity in the rotor frame
    T_body2inertial         = conditions.frames.body.transform_to_inertial
    T_inertial2body         = orientation_transpose(T_body2inertial)
    V_body                  = orientation_product(T_inertial2body,Vv)
    body2thrust,orientation = ducted_fan.body_to_prop_vel(commanded_TV) 
    T_body2thrust           = orientation_transpose(np.ones_like(T_body2inertial[:])*body2thrust)
    V_thrust                = orientation_product(T_body2thrust,V_body)

    # Check and correct for hover
    V         = V_thrust[:,0,None]
    V[V==0.0] = 1E-6
     
    tip_mach = (omega * ducted_fan.tip_radius) / a
    mach     =  V/ a
    # create tuple for querying surrogate 
    pts      = (mach,tip_mach,altitude) 
    
    thrust         = ducted_fan.performance_surrogates.thrust(pts)            
    power          = ducted_fan.performance_surrogates.power(pts)                 
    efficiency     = ducted_fan.performance_surrogates.efficiency(pts)            
    torque         = ducted_fan.performance_surrogates.torque(pts)                
    Ct             = ducted_fan.performance_surrogates.thrust_coefficient(pts)    
    Cp             = ducted_fan.performance_surrogates.power_coefficient(pts) 
    Cq             = torque/(rho*(n*n)*(D*D*D*D*D))
    FoM            = thrust*np.sqrt(thrust/(2*rho*A))/power  
    
    # calculate coefficients    
    thrust_prop_frame      = np.zeros((ctrl_pts,3))
    thrust_prop_frame[:,0] = thrust[:,0]
    thrust_vector          = orientation_product(orientation_transpose(T_body2thrust),thrust_prop_frame)
    
    conditions.energy.converters[ducted_fan.tag] = Data(
            torque                            = torque,
            thrust                            = thrust_vector,  
            power                             = power, 
            rpm                               = omega /Units.rpm ,   
            tip_mach                          = tip_mach, 
            efficiency                        = efficiency,         
            number_radial_stations            = Nr, 
            orientation                       = orientation, 
            speed_of_sound                    = conditions.freestream.speed_of_sound,
            density                           = conditions.freestream.density,
            velocity                          = Vv,     
            omega                             = omega,  
            thrust_per_blade                  = thrust/B,
            thrust_coefficient                = Ct, 
            torque_per_blade                  = torque/B,
            figure_of_merit                   = FoM, 
            torque_coefficient                = Cq,
            power_coefficient                 = Cp)
    
    return 


# RCAIDE/Methods/Powertrain/Sources/Batteries/Lithium_Ion_LFP/__init__.py
# 

"""
This module provides functionality for modeling lithium-ion LFP (lithium iron phosphate) batteries in powertrains. 
It includes methods for computing LFP cell performance and updating cell age.

See Also
--------
RCAIDE.Library.Components.Powertrain.Sources.Battery_Modules.Lithium_Ion_LFP
RCAIDE.Library.Methods.Powertrain.Sources.Batteries
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
from .compute_lfp_cell_performance   import * 
from .update_lfp_cell_age            import * 


# RCAIDE/Library/Methods/Emissions/Chemical_Reactor_Network_Method/evaluate_CRN_emission_indices.py
#  
# Created: Jul 2024, M. Clarke, M. Guidotti

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import  RCAIDE
from    RCAIDE.Framework.Core import Data
from    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_cantera import evaluate_cantera 
 
# package imports
import  numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  evaluate_correlation_emissions_indices
# ---------------------------------------------------------------------------------------------------------------------- 
def evaluate_CRN_emission_indices_no_surrogate(segment,settings,vehicle):

    """
    Computes emission indices directly using Chemical Reactor Network without surrogate models.

    Parameters
    ----------
    segment : Data
        Mission segment data container
            - state : Data
                Current state 
                    - numerics : Data
                        Numerical integration parameters
                            - time : Data
                                Time integration settings
                    - conditions : Data
                        Flight conditions and component states
                    - ones_row : function
                        Creates array of ones with specified size
                
    settings : Data
        Configuration settings for the simulation
        
    vehicle : Data
        Vehicle configuration data
            - networks : list
                List of propulsion system networks
                    - fuel_lines : list
                        Fuel distribution systems
                    - propulsors : list
                        Propulsion units 
    Returns
    -------
        Updates segment.state.conditions.emissions with:
            total : Data
                Total emissions over segment
                    - CO2 : float
                        Total CO2 emissions [kg]
                    - H2O : float
                        Total H2O emissions [kg]
                    - NOx : float
                        Total NOx emissions [kg]
            index : Data
                Emission indices
                    - CO2 : ndarray
                        CO2 emission index [kg_CO2/kg_fuel]
                    - CO : ndarray
                        CO emission index [kg_CO/kg_fuel]
                    - H2O : ndarray
                        H2O emission index [kg_H2O/kg_fuel]
                    - NO : ndarray
                        NO emission index [kg_NO/kg_fuel]
                    - NO2 : ndarray
                        NO2 emission index [kg_NO2/kg_fuel]

    Notes
    -----
    Computes emissions by directly evaluating the chemical kinetics at each time step
    using Cantera. 

    **Extra modules required**
        * numpy
        * Cantera (through evaluate_cantera function)

    See Also
    --------
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_cantera 

    References
    ----------
    [1] Goodwin, D. G., Speth, R. L., Moffat, H. K., & Weber, B. W. (2023). Cantera: An object-oriented software toolkit for chemical kinetics, thermodynamics, and transport processes (Version 3.0.0) [Computer software]. https://www.cantera.org
    """
  
    # unpack
    state     = segment.state
    I         = state.numerics.time.integrate
    
    CO2_total = 0 * state.ones_row(1)  
    CO_total  = 0 * state.ones_row(1) 
    H2O_total = 0 * state.ones_row(1) 
    NOx_total  = 0 * state.ones_row(1) 


    for network in vehicle.networks:  
        for fuel_line in network.fuel_lines:
            if fuel_line.active:  
                for p_i ,  propulsor in enumerate(network.propulsors):
                    if propulsor.active == True: 
                        if (type(propulsor) == RCAIDE.Library.Components.Powertrain.Propulsors.Turbofan) or \
                            type(propulsor) == RCAIDE.Library.Components.Powertrain.Converters.Turboshaft or \
                            type(propulsor) == RCAIDE.Library.Components.Powertrain.Propulsors.Turboprop or \
                            type(propulsor) == RCAIDE.Library.Components.Powertrain.Propulsors.Turbojet:    
                        
                            combustor = propulsor.combustor
                        
                            # unpack component conditions
                            n_cp                 = state.numerics.number_of_control_points 
                            propulsor_conditions = state.conditions.energy.propulsors[propulsor.tag] 
                            combustor_conditions = state.conditions.energy.converters[combustor.tag]  

                            T    = combustor_conditions.inputs.stagnation_temperature
                            P    = combustor_conditions.inputs.stagnation_pressure 
                            mdot = propulsor_conditions.core_mass_flow_rate 
                            FAR  = combustor_conditions.outputs.fuel_to_air_ratio 

                            mdot_fuel = propulsor_conditions.fuel_flow_rate

                            EI_CO2_comb    = 0 * state.ones_row(1)   
                            EI_CO_comb     = 0 * state.ones_row(1)  
                            EI_H2O_comb    = 0 * state.ones_row(1)  
                            EI_NOx_comb    = 0 * state.ones_row(1)  
                      

                            separate_zones = int(combustor.joint_mixing_fraction * combustor.N_SZ)
                            joint_zones    = int((1 - combustor.joint_mixing_fraction) * combustor.N_SZ)

                            combustor_PZ_phi           = np.zeros((n_cp,combustor.N_PZ))  # [-] 
                            combustor_PZ_T             = np.zeros((n_cp,combustor.N_PZ)) # [K] 
                            combustor_PZ_f_psr         = np.zeros((n_cp,combustor.N_PZ)) # [-] 
                            combustor_PZ_EI_CO2        = np.zeros((n_cp,combustor.N_PZ)) # [kg/kg_fuel]
                            combustor_PZ_EI_CO         = np.zeros((n_cp,combustor.N_PZ)) # [kg/kg_fuel]
                            combustor_PZ_EI_H2O        = np.zeros((n_cp,combustor.N_PZ)) # [kg/kg_fuel]
                            combustor_PZ_EI_NOx        = np.zeros((n_cp,combustor.N_PZ)) # [kg/kg_fuel]
                            combustor_SZ_sm_z          = np.zeros((n_cp,separate_zones)) # [-] 
                            combustor_SZ_sm_phi        = np.zeros((n_cp,separate_zones)) # [-] 
                            combustor_SZ_sm_T          = np.zeros((n_cp,separate_zones)) # [K] 
                            combustor_SZ_sm_EI_CO2     = np.zeros((n_cp,separate_zones)) # [kg/kg_fuel]
                            combustor_SZ_sm_EI_CO      = np.zeros((n_cp,separate_zones)) # [kg/kg_fuel]
                            combustor_SZ_sm_EI_H2O     = np.zeros((n_cp,separate_zones)) # [kg/kg_fuel]
                            combustor_SZ_sm_EI_NOx     = np.zeros((n_cp,separate_zones)) # [kg/kg_fuel]
                            combustor_SZ_fm_z          = np.zeros((n_cp,separate_zones)) # [-] 
                            combustor_SZ_fm_phi        = np.zeros((n_cp,separate_zones)) # [-] 
                            combustor_SZ_fm_T          = np.zeros((n_cp,separate_zones)) # [K] 
                            combustor_SZ_fm_EI_CO2     = np.zeros((n_cp,separate_zones)) # [kg/kg_fuel]
                            combustor_SZ_fm_EI_CO      = np.zeros((n_cp,separate_zones)) # [kg/kg_fuel]
                            combustor_SZ_fm_EI_H2O     = np.zeros((n_cp,separate_zones)) # [kg/kg_fuel]
                            combustor_SZ_fm_EI_NOx     = np.zeros((n_cp,separate_zones)) # [kg/kg_fuel]
                            combustor_SZ_joint_z       = np.zeros((n_cp,joint_zones)) # [-] 
                            combustor_SZ_joint_phi     = np.zeros((n_cp,joint_zones)) # [-] 
                            combustor_SZ_joint_T       = np.zeros((n_cp,joint_zones)) # [K] 
                            combustor_SZ_joint_EI_CO2  = np.zeros((n_cp,joint_zones)) # [kg/kg_fuel]
                            combustor_SZ_joint_EI_CO   = np.zeros((n_cp,joint_zones)) # [kg/kg_fuel]
                            combustor_SZ_joint_EI_H2O  = np.zeros((n_cp,joint_zones)) # [kg/kg_fuel]
                            combustor_SZ_joint_EI_NOx  = np.zeros((n_cp,joint_zones)) # [kg/kg_fuel]
                           
                            if network.identical_propulsors == True and p_i != 0:
                                EI_CO2_comb = EI_CO2_prev
                                EI_CO_comb  = EI_CO_prev
                                EI_H2O_comb = EI_H2O_prev
                                EI_NOx_comb  = EI_NOx_prev

                                combustor_PZ_phi           = combustor_PZ_phi_prev          
                                combustor_PZ_T             = combustor_PZ_T_prev            
                                combustor_PZ_f_psr         = combustor_PZ_f_psr_prev        
                                combustor_PZ_EI_CO2        = combustor_PZ_EI_CO2_prev       
                                combustor_PZ_EI_CO         = combustor_PZ_EI_CO_prev        
                                combustor_PZ_EI_H2O        = combustor_PZ_EI_H2O_prev       
                                combustor_PZ_EI_NOx        = combustor_PZ_EI_NOx_prev            
                                combustor_SZ_sm_z          = combustor_SZ_sm_z_prev         
                                combustor_SZ_sm_phi        = combustor_SZ_sm_phi_prev       
                                combustor_SZ_sm_T          = combustor_SZ_sm_T_prev         
                                combustor_SZ_sm_EI_CO2     = combustor_SZ_sm_EI_CO2_prev    
                                combustor_SZ_sm_EI_CO      = combustor_SZ_sm_EI_CO_prev     
                                combustor_SZ_sm_EI_H2O     = combustor_SZ_sm_EI_H2O_prev    
                                combustor_SZ_sm_EI_NOx     = combustor_SZ_sm_EI_NOx_prev      
                                combustor_SZ_fm_z          = combustor_SZ_fm_z_prev         
                                combustor_SZ_fm_phi        = combustor_SZ_fm_phi_prev       
                                combustor_SZ_fm_T          = combustor_SZ_fm_T_prev         
                                combustor_SZ_fm_EI_CO2     = combustor_SZ_fm_EI_CO2_prev    
                                combustor_SZ_fm_EI_CO      = combustor_SZ_fm_EI_CO_prev     
                                combustor_SZ_fm_EI_H2O     = combustor_SZ_fm_EI_H2O_prev    
                                combustor_SZ_fm_EI_NOx     = combustor_SZ_fm_EI_NOx_prev      
                                combustor_SZ_joint_z       = combustor_SZ_joint_z_prev      
                                combustor_SZ_joint_phi     = combustor_SZ_joint_phi_prev    
                                combustor_SZ_joint_T       = combustor_SZ_joint_T_prev      
                                combustor_SZ_joint_EI_CO2  = combustor_SZ_joint_EI_CO2_prev 
                                combustor_SZ_joint_EI_CO   = combustor_SZ_joint_EI_CO_prev  
                                combustor_SZ_joint_EI_H2O  = combustor_SZ_joint_EI_H2O_prev 
                                combustor_SZ_joint_EI_NOx  = combustor_SZ_joint_EI_NOx_prev 
                                
                            else:     
                                for t_idx in range(n_cp):
                                    # Call cantera 
                                    results = evaluate_cantera(combustor,T[t_idx,0],P[t_idx,0],mdot[t_idx,0],FAR[t_idx,0])
                                    
                                    EI_CO2_comb[t_idx,0]                = results.final.EI.CO2
                                    EI_CO_comb[t_idx,0]                 = results.final.EI.CO 
                                    EI_H2O_comb[t_idx,0]                = results.final.EI.H2O
                                    EI_NOx_comb[t_idx,0]                = results.final.EI.NOx 
                                    combustor_PZ_phi[t_idx,:]           = results.PZ.phi
                                    combustor_PZ_T[t_idx,:]             = results.PZ.T
                                    combustor_PZ_f_psr[t_idx,:]         = results.PZ.f_psr
                                    combustor_PZ_EI_CO2[t_idx,:]        = results.PZ.EI.CO2
                                    combustor_PZ_EI_CO[t_idx,:]         = results.PZ.EI.CO
                                    combustor_PZ_EI_H2O[t_idx,:]        = results.PZ.EI.H2O
                                    combustor_PZ_EI_NOx[t_idx,:]        = results.PZ.EI.NOx
                                    combustor_SZ_sm_z[t_idx,:]          = results.SZ.sm.z
                                    combustor_SZ_sm_phi[t_idx,:]        = results.SZ.sm.phi
                                    combustor_SZ_sm_T[t_idx,:]          = results.SZ.sm.T
                                    combustor_SZ_sm_EI_CO2[t_idx,:]     = results.SZ.sm.EI.CO2
                                    combustor_SZ_sm_EI_CO[t_idx,:]      = results.SZ.sm.EI.CO
                                    combustor_SZ_sm_EI_H2O[t_idx,:]     = results.SZ.sm.EI.H2O
                                    combustor_SZ_sm_EI_NOx[t_idx,:]     = results.SZ.sm.EI.NOx
                                    combustor_SZ_fm_z[t_idx,:]          = results.SZ.fm.z
                                    combustor_SZ_fm_phi[t_idx,:]        = results.SZ.fm.phi
                                    combustor_SZ_fm_T[t_idx,:]          = results.SZ.fm.T
                                    combustor_SZ_fm_EI_CO2[t_idx,:]     = results.SZ.fm.EI.CO2          
                                    combustor_SZ_fm_EI_CO[t_idx,:]      = results.SZ.fm.EI.CO     
                                    combustor_SZ_fm_EI_H2O[t_idx,:]     = results.SZ.fm.EI.H2O
                                    combustor_SZ_fm_EI_NOx[t_idx,:]     = results.SZ.fm.EI.NOx
                                    combustor_SZ_joint_z[t_idx,:]       = results.SZ.joint.z
                                    combustor_SZ_joint_phi[t_idx,:]     = results.SZ.joint.phi
                                    combustor_SZ_joint_T[t_idx,:]       = results.SZ.joint.T
                                    combustor_SZ_joint_EI_CO2[t_idx,:]  = results.SZ.joint.EI.CO2
                                    combustor_SZ_joint_EI_CO[t_idx,:]   = results.SZ.joint.EI.CO  
                                    combustor_SZ_joint_EI_H2O[t_idx,:]  = results.SZ.joint.EI.H2O 
                                    combustor_SZ_joint_EI_NOx[t_idx,:]  = results.SZ.joint.EI.NOx 
                                    
                                    EI_CO2_prev                     = EI_CO2_comb 
                                    EI_CO_prev                      = EI_CO_comb  
                                    EI_H2O_prev                     = EI_H2O_comb 
                                    EI_NOx_prev                     = EI_NOx_comb  
                                    combustor_PZ_phi_prev           = combustor_PZ_phi          
                                    combustor_PZ_T_prev             = combustor_PZ_T            
                                    combustor_PZ_f_psr_prev         = combustor_PZ_f_psr        
                                    combustor_PZ_EI_CO2_prev        = combustor_PZ_EI_CO2       
                                    combustor_PZ_EI_CO_prev         = combustor_PZ_EI_CO        
                                    combustor_PZ_EI_H2O_prev        = combustor_PZ_EI_H2O       
                                    combustor_PZ_EI_NOx_prev        = combustor_PZ_EI_NOx       
                                    combustor_SZ_sm_z_prev          = combustor_SZ_sm_z         
                                    combustor_SZ_sm_phi_prev        = combustor_SZ_sm_phi       
                                    combustor_SZ_sm_T_prev          = combustor_SZ_sm_T         
                                    combustor_SZ_sm_EI_CO2_prev     = combustor_SZ_sm_EI_CO2    
                                    combustor_SZ_sm_EI_CO_prev      = combustor_SZ_sm_EI_CO     
                                    combustor_SZ_sm_EI_H2O_prev     = combustor_SZ_sm_EI_H2O    
                                    combustor_SZ_sm_EI_NOx_prev     = combustor_SZ_sm_EI_NOx    
                                    combustor_SZ_fm_z_prev          = combustor_SZ_fm_z         
                                    combustor_SZ_fm_phi_prev        = combustor_SZ_fm_phi       
                                    combustor_SZ_fm_T_prev          = combustor_SZ_fm_T         
                                    combustor_SZ_fm_EI_CO2_prev     = combustor_SZ_fm_EI_CO2    
                                    combustor_SZ_fm_EI_CO_prev      = combustor_SZ_fm_EI_CO     
                                    combustor_SZ_fm_EI_H2O_prev     = combustor_SZ_fm_EI_H2O    
                                    combustor_SZ_fm_EI_NOx_prev     = combustor_SZ_fm_EI_NOx     
                                    combustor_SZ_joint_z_prev       = combustor_SZ_joint_z      
                                    combustor_SZ_joint_phi_prev     = combustor_SZ_joint_phi    
                                    combustor_SZ_joint_T_prev       = combustor_SZ_joint_T      
                                    combustor_SZ_joint_EI_CO2_prev  = combustor_SZ_joint_EI_CO2 
                                    combustor_SZ_joint_EI_CO_prev   = combustor_SZ_joint_EI_CO  
                                    combustor_SZ_joint_EI_H2O_prev  = combustor_SZ_joint_EI_H2O 
                                    combustor_SZ_joint_EI_NOx_prev  = combustor_SZ_joint_EI_NOx 
                                
                            CO2_total  += np.dot(I,mdot_fuel*EI_CO2_comb)
                            CO_total   += np.dot(I,mdot_fuel *EI_CO_comb )
                            H2O_total  += np.dot(I,mdot_fuel*EI_H2O_comb)
                            NOx_total   += np.dot(I,mdot_fuel *EI_NOx_comb ) 
                           

    emissions                        = Data()
    emissions.total                  = Data()
    emissions.index                  = Data() 
    emissions.total.CO2              = CO2_total  * combustor.fuel_data.global_warming_potential_100.CO2 
    emissions.total.H2O              = H2O_total  * combustor.fuel_data.global_warming_potential_100.H2O  
    emissions.total.NOx              = NOx_total * combustor.fuel_data.global_warming_potential_100.NOx 
    emissions.index.CO2              = EI_CO2_comb
    emissions.index.CO               = EI_CO_comb 
    emissions.index.H2O              = EI_H2O_comb
    emissions.index.NOx              = EI_NOx_comb 
    emissions.index.PZ_phi           = combustor_PZ_phi
    emissions.index.PZ_T             = combustor_PZ_T
    emissions.index.PZ_f_psr         = combustor_PZ_f_psr
    emissions.index.PZ_EI_CO2        = combustor_PZ_EI_CO2
    emissions.index.PZ_EI_CO         = combustor_PZ_EI_CO
    emissions.index.PZ_EI_H2O        = combustor_PZ_EI_H2O
    emissions.index.PZ_EI_NOx        = combustor_PZ_EI_NOx
    emissions.index.SZ_sm_z          = combustor_SZ_sm_z
    emissions.index.SZ_sm_phi        = combustor_SZ_sm_phi
    emissions.index.SZ_sm_T          = combustor_SZ_sm_T   
    emissions.index.SZ_sm_EI_CO2     = combustor_SZ_sm_EI_CO2
    emissions.index.SZ_sm_EI_CO      = combustor_SZ_sm_EI_CO
    emissions.index.SZ_sm_EI_H2O     = combustor_SZ_sm_EI_H2O
    emissions.index.SZ_sm_EI_NOx     = combustor_SZ_sm_EI_NOx
    emissions.index.SZ_fm_z          = combustor_SZ_fm_z
    emissions.index.SZ_fm_phi        = combustor_SZ_fm_phi
    emissions.index.SZ_fm_T          = combustor_SZ_fm_T
    emissions.index.SZ_fm_EI_CO2     = combustor_SZ_fm_EI_CO2
    emissions.index.SZ_fm_EI_CO      = combustor_SZ_fm_EI_CO
    emissions.index.SZ_fm_EI_H2O     = combustor_SZ_fm_EI_H2O
    emissions.index.SZ_fm_EI_NOx     = combustor_SZ_fm_EI_NOx
    emissions.index.SZ_joint_z       = combustor_SZ_joint_z
    emissions.index.SZ_joint_phi     = combustor_SZ_joint_phi
    emissions.index.SZ_joint_T       = combustor_SZ_joint_T
    emissions.index.SZ_joint_EI_CO2  = combustor_SZ_joint_EI_CO2
    emissions.index.SZ_joint_EI_CO   = combustor_SZ_joint_EI_CO
    emissions.index.SZ_joint_EI_H2O  = combustor_SZ_joint_EI_H2O
    emissions.index.SZ_joint_EI_NOx  = combustor_SZ_joint_EI_NOx
        
    state.conditions.emissions =  emissions
    return   
    

def evaluate_CRN_emission_indices_surrogate(segment,settings,vehicle): 

    """
    Computes emission indices using pre-trained Chemical Reactor Network surrogate models.

    Parameters
    ----------
    segment : Data
        Mission segment data container
        
        - state : Data
            Current state of the system

            - numerics : Data
                Numerical integration parameters
            - conditions : Data
                Flight conditions and component states
        - analyses : Data
            Analysis settings and models

            - emissions : Data
                Emissions analysis settings

                - surrogates : Data
                    Trained surrogate models for each species
                
    settings : Data
        Configuration settings for the simulation
        
    vehicle : Data
        Vehicle configuration data

        - networks : list
            List of propulsion system networks

            - propulsors : list
                Propulsion units 

    Returns
    -------
    Updates segment.state.conditions.emissions with:
        
    total : Data
        Total emissions over segment

        - CO2 : float
            Total CO2 emissions [kg]
        - H2O : float
            Total H2O emissions [kg]
        - NOx : float
            Total NOx emissions [kg]
    index : Data
        Emission indices

        - CO2 : ndarray
            CO2 emission index [kg_CO2/kg_fuel]
        - CO : ndarray
            CO emission index [kg_CO/kg_fuel]
        - H2O : ndarray
            H2O emission index [kg_H2O/kg_fuel]
        - NO : ndarray
            NO emission index [kg_NO/kg_fuel]
        - NO2 : ndarray
            NO2 emission index [kg_NO2/kg_fuel]

    Notes
    -----
    Uses pre-trained surrogate models to estimate emission indices.

    **Extra modules required**

    * numpy

    **Major Assumptions**

    * Operating conditions fall within the training data range
    * Linear interpolation can be employed between training points

    See Also
    --------
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.train_CRN_EI_surrogates
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.build_CRN_EI_surrogates
    RCAIDE.Library.Methods.Emissions.Chemical_Reactor_Network_Method.evaluate_cantera 

    References
    ----------
    [1] Goodwin, D. G., Speth, R. L., Moffat, H. K., & Weber, B. W. (2023). Cantera: An object-oriented software toolkit for chemical kinetics, thermodynamics, and transport processes (Version 3.0.0) [Computer software]. https://www.cantera.org
    """
  
    I          = segment.state.numerics.time.integrate
    surrogates = segment.analyses.emissions.surrogates
    
    CO2_total = 0 * segment.state.ones_row(1)  
    CO_total  = 0 * segment.state.ones_row(1) 
    H2O_total = 0 * segment.state.ones_row(1) 
    NOx_total  = 0 * segment.state.ones_row(1) 


    for network in vehicle.networks:    
        for propulsor in network.propulsors:
            if propulsor.active == True:
                if (type(propulsor) == RCAIDE.Library.Components.Powertrain.Propulsors.Turbofan) or \
                    type(propulsor) == RCAIDE.Library.Components.Powertrain.Propulsors.Turboprop or \
                    type(propulsor) == RCAIDE.Library.Components.Powertrain.Converters.Turboshaft or \
                    type(propulsor) == RCAIDE.Library.Components.Powertrain.Propulsors.Turbojet:    
                
                    combustor = propulsor.combustor
                
                    # unpack component conditions
                    propulsor_conditions = segment.state.conditions.energy.propulsors[propulsor.tag] 
                    combustor_conditions = segment.state.conditions.energy.converters[combustor.tag]  

                    T    = combustor_conditions.inputs.stagnation_temperature
                    P    = combustor_conditions.inputs.stagnation_pressure 
                    mdot = propulsor_conditions.core_mass_flow_rate 
                    FAR  = combustor_conditions.outputs.fuel_to_air_ratio 

                    mdot_fuel = propulsor_conditions.fuel_flow_rate
                    
                    pts = np.hstack((T,P,mdot,FAR)) 

                    EI_CO2_comb  = np.atleast_2d(surrogates.EI_CO2(pts)).T
                    EI_CO_comb   = np.atleast_2d(surrogates.EI_CO(pts)).T 
                    EI_H2O_comb  = np.atleast_2d(surrogates.EI_H2O(pts)).T 
                    EI_NOx_comb  = np.atleast_2d(surrogates.EI_NOx(pts)).T 
                          
                    CO2_total += np.dot(I,mdot_fuel*EI_CO2_comb)
                    CO_total  += np.dot(I,mdot_fuel *EI_CO_comb )
                    H2O_total += np.dot(I,mdot_fuel*EI_H2O_comb)
                    NOx_total  += np.dot(I,mdot_fuel *EI_NOx_comb ) 


    emissions                 = Data()
    emissions.total           = Data()
    emissions.index           = Data() 
    emissions.total.CO2       = CO2_total * combustor.fuel_data.global_warming_potential_100.CO2 
    emissions.total.H2O       = H2O_total * combustor.fuel_data.global_warming_potential_100.H2O  
    emissions.total.NOx       = NOx_total * combustor.fuel_data.global_warming_potential_100.NOx 
    emissions.index.CO2       = EI_CO2_comb
    emissions.index.CO        = EI_CO_comb 
    emissions.index.H2O       = EI_H2O_comb
    emissions.index.NOx       = EI_NOx_comb 
 
    segment.state.conditions.emissions =  emissions
    return   

     


# RCAIDE/Library/Missions/Segments/Descent/Constant_CAS_Constant_Rate.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------  
#  IMPORT 
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core import Units

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  
def initialize_conditions(segment):
    """
    Initializes conditions for constant calibrated airspeed descent at fixed rate

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - descent_rate : float
                Rate of descent [m/s]
            - calibrated_air_speed : float
                Calibrated airspeed to maintain [m/s]
            - altitude_start : float
                Initial altitude [m]
            - altitude_end : float
                Final altitude [m]
            - sideslip_angle : float
                Aircraft sideslip angle [rad]
            - temperature_deviation : float
                Temperature offset from standard day [K]
            - state:
                numerics.dimensionless.control_points : array
                    Discretization points [-]
                conditions : Data
                    State conditions container
            - analyses:
                atmosphere : Model
                    Atmospheric model for property calculations
    
    Returns
    -------
    None

    Notes
    -----
    This function sets up the initial conditions for a descent segment with constant
    calibrated airspeed (CAS) and constant descent rate. It handles the conversion
    between CAS and true airspeed accounting for pressure and density variations
    with altitude. Updates segment conditions directly with velocity_vector [m/s], altitude [m],
    and position_vector [m].

    **Calculation Process**
        1. Discretize altitude profile
        2. Get atmospheric properties at each altitude
        3. Convert CAS to true airspeed using:
            - Pressure ratio (δ)
            - Compressibility effects
            - Equivalent airspeed (EAS) conversion
        4. Decompose velocity into components using:
            - Fixed descent rate
            - Sideslip angle
            - Computed true airspeed

    **Major Assumptions**
        * Constant calibrated airspeed
        * Constant descent rate
        * Standard atmosphere model with temperature deviation
        * Small angle approximations
        * Quasi-steady flight

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    RCAIDE.Library.Mission.Common.Update.atmosphere
    """       
    
    # unpack
    descent_rate = segment.descent_rate
    cas          = segment.calibrated_air_speed   
    alt0         = segment.altitude_start 
    altf         = segment.altitude_end
    beta         = segment.sideslip_angle
    t_nondim     = segment.state.numerics.dimensionless.control_points
    conditions   = segment.state.conditions  

    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]

    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0
    
    # determine airspeed from calibrated airspeed
    RCAIDE.Library.Mission.Common.Update.atmosphere(segment) # get density for airspeed

    alt_data = segment.analyses.atmosphere.compute_values(alt,segment.temperature_deviation)
    density  = alt_data.density[:,0]  
    pressure = alt_data.pressure[:,0] 

    MSL_data  = segment.analyses.atmosphere.compute_values(0.0,segment.temperature_deviation)
    pressure0 = MSL_data.pressure[0]
    

    if cas is None:
        if not segment.state.initials: raise AttributeError('initial equivalent airspeed not set')
        air_speed =  np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1,:])    
    else:  
        kcas  = cas / Units.knots
        delta = pressure / pressure0 
    
        mach = 2.236*((((1+4.575e-7*kcas**2)**3.5-1)/delta + 1)**0.2857 - 1)**0.5
    
        qc  = pressure * ((1+0.2*mach**2)**3.5 - 1)
        eas = cas * (pressure/pressure0)**0.5*(((qc/pressure+1)**0.286-1)/((qc/pressure0+1)**0.286-1))**0.5
        
        air_speed = eas/np.sqrt(density/MSL_data.density[0])    
    
    # process velocity vector
    v_mag = air_speed
    v_z   = descent_rate # z points down
    v_xy  = np.sqrt( v_mag**2 - v_z**2 )
    v_x   = np.cos(beta)*v_xy
    v_y   = np.sin(beta)*v_xy
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x
    conditions.frames.inertial.velocity_vector[:,1] = v_y
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude in this context


# RCAIDE/Library/Missions/Segments/Ground/Battery_Charge_Discharge.py
# 
# 
# Created:  Jul 2023, M. Clarke 
import RCAIDE 
from RCAIDE.Framework.Core import  Units
import  numpy as  np 
# ----------------------------------------------------------------------------------------------------------------------  
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------    
def initialize_conditions(segment):  
    """
    Initializes conditions for battery charging or discharging on ground

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - cutoff_SOC : float
                Target state of charge to reach [-]
            - cooling_time : float
                Additional time for battery cooling [s]
            - initial_battery_state_of_charge : float, optional
                Initial SOC if no previous segment [-]
            - analyses:
                energy:
                    vehicle:
                        networks : list
                            List of power networks
                            Each network contains:
                            - busses : list
                                List of electrical busses
                                Each bus contains:
                                - charging_c_rate : float
                                    Battery charging C-rate [-]
                                - battery_modules : list
                                    List of battery modules
        
        For discharging segments:
            - time : float
                Duration of discharge [s]
    
    Returns
    -------
    None
        Updates segment conditions directly:
            - conditions.frames.inertial.time [s]
    
    Notes
    -----
    This function sets up the initial conditions for a ground segment involving 
    battery charging or discharging. For charging segments, the charging time is 
    determined by the battery with the lowest state of charge.

    **Calculation Process**
    For charging:
        1. Find lowest battery state of charge
        2. Calculate charging time based on:
            t = (SOC_target - SOC_current) / C_rate
        3. Add cooling time
        4. Discretize time points

    For discharging:
        1. Use specified discharge time
        2. Discretize time points

    **Major Assumptions**
        * Charging time determined by lowest SOC battery
        * Constant C-rate charging
        * Linear SOC increase during charging
        * Uniform cooling time added

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    """    
    t_nondim   = segment.state.numerics.dimensionless.control_points

    if isinstance(segment, RCAIDE.Framework.Mission.Segments.Ground.Battery_Recharge):
        for network in segment.analyses.energy.vehicle.networks:
            time =  []
            for bus in  network.busses:
                t=0
                if segment.state.initials.keys():
                    end_of_flight_soc = 1
                    for battery_module in segment.state.conditions.energy[bus.tag].battery_modules:
                        end_of_flight_soc = min(end_of_flight_soc,battery_module.cell.state_of_charge[-1])
                else:
                    end_of_flight_soc =  segment.initial_battery_state_of_charge
                
                t           =  max(((segment.cutoff_SOC-end_of_flight_soc) / bus.charging_c_rate )*Units.hrs  , t) 
                t           += segment.cooling_time
                time.append(t)
            t_initial = segment.state.conditions.frames.inertial.time[0,0]
            t_nondim  = segment.state.numerics.dimensionless.control_points
            time      = np.max(time)
            charging_time      = t_nondim * ( time ) + t_initial 
            segment.state.conditions.frames.inertial.time[:,0] = charging_time[:,0]

    else:

        t_initial = segment.state.conditions.frames.inertial.time[0,0]
        t_nondim  = segment.state.numerics.dimensionless.control_points
        time      = t_nondim * ( segment.time ) + t_initial
        segment.state.conditions.frames.inertial.time[:,0] = time[:,0]


# RCAIDE/Methods/Mission/Segments/Cruise/__init__.py
# 

"""
Collection of cruise segment types for aircraft mission analysis. This module provides various 
cruise profiles including constant Mach, constant speed, loiter patterns, and specialized 
maneuvers at constant altitude with different control parameters (throttle, dynamic pressure, 
pitch rate, etc.).

See Also
--------
RCAIDE.Library.Mission.Segments.Climb
RCAIDE.Library.Mission.Segments.Descent
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from . import Constant_Mach_Constant_Altitude
from . import Constant_Speed_Constant_Altitude
from . import Constant_Mach_Constant_Altitude_Loiter
from . import Constant_Throttle_Constant_Altitude 
from . import Constant_Dynamic_Pressure_Constant_Altitude_Loiter
from . import Constant_Acceleration_Constant_Altitude
from . import Constant_Pitch_Rate_Constant_Altitude
from . import Constant_Dynamic_Pressure_Constant_Altitude
from . import Constant_Speed_Constant_Altitude_Loiter
from . import Curved_Constant_Radius_Constant_Speed_Constant_Altitude



# RCAIDE/Library/Missions/Common/Initialize/time.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Time
# ---------------------------------------------------------------------------------------------------------------------- 
def time(segment):
    """
    Initializes time variables for mission segment analysis

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed

    Notes
    -----
    This function sets up the initial time values for both inertial and
    planetary reference frames. It handles time continuity between segments
    and establishes start times for new segments.

    The function follows this priority for time initialization:
        1. Previous segment final time (if initials exist)
        2. Explicit segment start time (if specified)
        3. Current initial time value

    **Required Segment State Variables**

    If segment.state.initials exists:
        state.initials.conditions.frames:
            inertial:
                - time : array
                    Previous segment final time [s]
            planet:
                - start_time : float
                    Previous segment start time [s]

    state.conditions.frames:
        inertial:
            - time : array
                Current segment time array [s]
        planet:
            - start_time : float
                Current segment start time [s]

    **Optional Segment Parameters**
    
    segment:
        - start_time : float
            Explicit segment start time [s]

    **Major Assumptions**
        * Continuous time tracking when using initials
        * Valid time values (non-negative)
        * Proper time synchronization between frames
        * Time measured in seconds

    Returns
    -------
    None
        Updates segment conditions directly

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    RCAIDE.Library.Mission.Common.Initialize.planet_position
    """        
    
    if segment.state.initials:
        t_initial = segment.state.initials.conditions.frames.inertial.time
        t_current = segment.state.conditions.frames.inertial.time 
        segment.state.conditions.frames.inertial.time[:,:] = t_current + (t_initial[-1,0] - t_current[0,0])
        
    else:
        t_initial = segment.state.conditions.frames.inertial.time[0,0]
        
    if segment.state.initials:
        segment.state.conditions.frames.planet.start_time = segment.state.initials.conditions.frames.planet.start_time
        
    elif 'start_time' in segment:
        segment.state.conditions.frames.planet.start_time = segment.start_time
    
    return 



# RCAIDE/Core/Container.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------        
import RCAIDE
from .               import Data
from warnings        import warn
import random

import string
chars = string.punctuation + string.whitespace
t_table = str.maketrans( chars          + string.ascii_uppercase , 
                            '_'*len(chars) + string.ascii_lowercase )

# ----------------------------------------------------------------------------------------------------------------------
#  Container
# ----------------------------------------------------------------------------------------------------------------------   

class Container(Data):
    """ A dict-type container with attribute, item and index style access
        intended to hold a attribute-accessible list of Data(). This is unordered.
        
        Assumptions:
        N/A
        
        Source:
        N/A
        
    """
            
        
    def __defaults__(self):
        """ Defaults function
    
            Assumptions:
            None
        
            Source:
            N/A
        
            Inputs:
            N/A
        
            Outputs:
            N/A
            
            Properties Used:
            N/A
        """          
        pass
    
    def __init__(self,*args,**kwarg):
        """ Initialization that builds the container
        
            Assumptions:
            None
        
            Source:
            N/A
        
            Inputs:
            self
        
            Outputs:
            N/A
            
            Properties Used:
            N/A
        """          
        super(Container,self).__init__(*args,**kwarg)
        self.__defaults__()
    
    def append(self,val):
        """ Appends the value to the containers
            This overrides the Data class append by allowing for duplicate named components
            The following components will get new names.
        
            Assumptions:
            None
        
            Source:
            N/A
        
            Inputs:
            self
        
            Outputs:
            N/A
            
            Properties Used:
            N/A
        """           
        
        old_tags = []
        old_tags = get_tags(self,old_tags) 
        check_tags(val,old_tags)     
        Data.append(self,val) 
         
        return
    
        
    def extend(self,vals):
        # """ Append things regressively depending on what is inside.
    
        #     Assumptions:
        #     None
        
        #     Source:
        #     N/A
        
        #     Inputs:
        #     self
        
        #     Outputs:
        #     N/A
            
        #     Properties Used:
        #     N/A
        # """         
        if isinstance(vals,(list,tuple)):
            for v in val: self.append(v)
        elif isinstance(vals,dict):
            self.update(vals)
        else:
            raise Exception('unrecognized data type') 

def get_tags(item,tag_list):
        
    if isinstance(item, RCAIDE.Library.Components.Component) or isinstance(item, dict):
        for s_tag, s_item in item.items():
            if 'tag' == s_tag:
                item.tag = str.lower(s_item.translate(t_table))
                tag_list.append(item.tag)
            if isinstance(s_item, RCAIDE.Library.Components.Component): 
                for ss_tag, ss_item in s_item.items(): 
                    if 'tag' == ss_tag:
                        s_item.tag = str.lower(ss_item.translate(t_table))
                        tag_list.append(s_item.tag)
                    if isinstance(ss_item, RCAIDE.Library.Components.Component):
                        for sss_tag, sss_item in ss_item.items():
                            if 'tag' == sss_tag:
                                ss_item.tag = str.lower(sss_item.translate(t_table))
                                tag_list.append(ss_item.tag) 
                            if isinstance(sss_item, RCAIDE.Library.Components.Component):
                                for ssss_tag, ssss_item in sss_item.items():
                                    if 'tag' == ssss_tag:
                                        sss_item.tag = str.lower(ssss_item.translate(t_table))
                                        tag_list.append(sss_item.tag)                                
                            
    return tag_list
                            
def check_tags(item,tag_list):
    if isinstance(item, RCAIDE.Library.Components.Component) or isinstance(item, dict):
        for s_tag, s_item in item.items():
            if 'tag' == s_tag:
                if s_item in tag_list:
                    unmodified_tag = str.lower(s_item.translate(t_table))
                    string_of_keys = "".join(tag_list)
                    n_comps        = string_of_keys.count(unmodified_tag)
                    item.tag       = unmodified_tag + str(n_comps+1)
                else:
                    item.tag = str.lower(s_item.translate(t_table))
                tag_list.append(item.tag)
            if isinstance(s_item, RCAIDE.Library.Components.Component): 
                for ss_tag, ss_item in s_item.items(): 
                    if 'tag' == ss_tag: 
                        if ss_item in tag_list:
                            unmodified_tag = str.lower(ss_item.translate(t_table))  
                            string_of_keys = "".join(tag_list)
                            n_comps        = string_of_keys.count(unmodified_tag)
                            s_item.tag     = unmodified_tag + str(n_comps+1)
                        else:
                            s_item.tag = str.lower(ss_item.translate(t_table))
                        tag_list.append(s_item.tag) 
                    if isinstance(ss_item, RCAIDE.Library.Components.Component):
                        for sss_tag, sss_item in ss_item.items():
                            if 'tag' == sss_tag: 
                                if sss_item in tag_list:
                                    unmodified_tag = str.lower(sss_item.translate(t_table))  
                                    string_of_keys = "".join(tag_list)
                                    n_comps        = string_of_keys.count(unmodified_tag)
                                    ss_item.tag    = unmodified_tag + str(n_comps+1)
                                else:
                                    ss_item.tag = str.lower(sss_item.translate(t_table))  
                                tag_list.append(ss_item.tag) 
                            if isinstance(sss_item, RCAIDE.Library.Components.Component):
                                for ssss_tag, ssss_item in sss_item.items():
                                    if 'tag' == ssss_tag: 
                                        if ssss_item in tag_list:
                                            unmodified_tag = str.lower(ssss_item.translate(t_table))  
                                            string_of_keys = "".join(tag_list)
                                            n_comps        = string_of_keys.count(unmodified_tag)
                                            sss_item.tag   = unmodified_tag + str(n_comps+1)
                                        else:
                                            sss_item.tag = str.lower(ssss_item.translate(t_table))  
                                        tag_list.append(sss_item.tag)                                   
    return  



# RCAIDE/Library/Missions/Segments/Climb/Linear_Mach_Constant_Rate.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE 
from RCAIDE.Library.Mission.Common.Update.atmosphere import atmosphere

# Package imports  
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    
    """
    Initializes conditions for linear Mach number climb segment

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed

    Notes
    -----
    This function sets up the initial conditions for a climb segment with linearly
    varying Mach number and constant rate of climb. The Mach number varies linearly
    between the start and end values.

    **Required Segment Components**

    segment:
        - climb_rate : float
            Rate of climb [m/s]
        - mach_number_start : float
            Initial Mach number [-]
        - mach_number_end : float
            Final Mach number [-]
        - altitude_start : float
            Initial altitude [m]
        - altitude_end : float
            Final altitude [m]
        - sideslip_angle : float
            Aircraft sideslip angle [rad]
        - state:
            numerics.dimensionless.control_points : array
                Discretization points [-]
            conditions : Data
                State conditions container
        - analyses:
            atmosphere : Model
                Atmospheric model for property calculations

    **Calculation Process**
        1. Discretize altitude profile
        2. Get atmospheric properties for speed of sound
        3. Calculate velocity from Mach number variation
        4. Decompose velocity into components using:
            - Climb rate constraint
            - Sideslip angle
            - Linear Mach profile

    **Major Assumptions**
        * Linear Mach number variation
        * Constant rate of climb
        * Standard atmosphere model
        * Small angle approximations
        * Quasi-steady flight

    Returns
    -------
    None
        Updates segment conditions directly:
        - conditions.frames.inertial.velocity_vector [m/s]
        - conditions.frames.inertial.position_vector [m]
        - conditions.freestream.altitude [m]

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    RCAIDE.Library.Mission.Common.Update.atmosphere
    """          
    
    # unpack
    climb_rate = segment.climb_rate
    M0         = segment.mach_number_start
    Mf         = segment.mach_number_end
    alt0       = segment.altitude_start 
    altf       = segment.altitude_end 
    beta       = segment.sideslip_angle
    t_nondim   = segment.state.numerics.dimensionless.control_points
    conditions = segment.state.conditions
    
    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]

    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude in this context
    

    # Update freestream to get speed of sound
    atmosphere(segment)
    a          = conditions.freestream.speed_of_sound
    
    # check for initial velocity
    if M0 is None: 
        if not segment.state.initials: raise AttributeError('airspeed not set')
        M0 = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])/ a
         
    # process velocity vector
    mach_number = (Mf-M0)*t_nondim + M0
    v_mag       = mach_number * a
    v_z         = -climb_rate # z points down
    v_xy        = np.sqrt( v_mag**2 - v_z**2 )
    v_x         = np.cos(beta)*v_xy
    v_y         = np.sin(beta)*v_xy
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x[:,0]
    conditions.frames.inertial.velocity_vector[:,1] = v_y[:,0]
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude in this context
    


# RCAIDE/Library/Missions/Segments/Common/Update/dimensionless.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Update Differentials Altitude
# ---------------------------------------------------------------------------------------------------------------------- 
def differentials_altitude(segment):
    """
    Initializes the differential altitude for mission segment time calculations

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed

    Notes
    -----
    This function calculates the time differentials based on altitude changes
    in the mission segment. It uses the vertical component of velocity and
    position to determine appropriate time steps for integration.

    The function performs the following steps:
        1. Extracts control points and integration operators
        2. Calculates total altitude change
        3. Determines time step based on vertical velocity
        4. Rescales time operators
        5. Updates segment time vector

    **Required Segment State Variables**

    state.numerics.dimensionless:
        - integrate : array
            Integration operator matrix
        - control_points : array
            Normalized time points

    state.conditions.frames.inertial:
        - position_vector : array
            Vehicle position in inertial frame [m]
        - velocity_vector : array
            Vehicle velocity in inertial frame [m/s]
        - time : array
            Segment time vector [s]

    **Major Assumptions**
        * Continuous vertical velocity
        * Well-defined altitude change
        * No singularities in vertical velocity

    Returns
    -------
    None
        Updates segment state directly

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    """

    # unpack
    t = segment.state.numerics.dimensionless.control_points 
    I = segment.state.numerics.dimensionless.integrate
    r = segment.state.conditions.frames.inertial.position_vector
    v = segment.state.conditions.frames.inertial.velocity_vector

    dz = r[-1,2] - r[0,2]
    vz = v[:,2,None] # maintain column array

    # get overall time step
    dt = np.dot( I[-1,:] * dz , 1/ vz[:,0] )

    # rescale operators
    t = t * dt

    # pack
    t_initial = segment.state.conditions.frames.inertial.time[0,0]
    segment.state.conditions.frames.inertial.time[:,0] = t_initial + t[:,0]

    return


# RCAIDE/Library/Components/Powertrain/Energy/Sources/Cryogenic_Tanks/__init__.py
# 

"""
Cryogenic tank components for aircraft cyrogen storage and distribution

This module provides implementations for various aircraft cryogenic tank types.
Each tank type has specific characteristics for cyrogen storage, weight distribution,
and system integration.

See Also
--------
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Cryogenic_Tank         import Cryogenic_Tank 


# RCAIDE/Library/Missions/Segments/converge.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
import RCAIDE
from RCAIDE.Framework.Core import  Units, Data
from RCAIDE.Framework.Optimization.Packages.scipy import scipy_setup
from RCAIDE.Framework.Optimization.Common         import Nexus
from RCAIDE.Framework.Analyses.Process            import Process

import scipy 
import scipy.optimize
import numpy as np 
import sys 
import os 


# ----------------------------------------------------------------------------------------------------------------------
# converge root
# ---------------------------------------------------------------------------------------------------------------------- 
def converge(segment):
    """Interfaces the mission a root finder algorithm.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    segment                            [Data]
    segment.settings.root_finder       [Data]
    state.numerics.tolerance_solution  [Unitless]

    Outputs:
    state.unknowns                     [Any]
    segment.state.numerics.converged   [Unitless]

    Properties Used:
    N/A
    """ 
    
    if segment.state.numerics.solver.type  == "optimize": 
        problem  = add_mission_variables(segment) 
       
        # Commense suppression of console window output  
        devnull = open(os.devnull,'w')
        sys.stdout = devnull
         
        outputs  = scipy_setup.SciPy_Solve(problem,
                                           solver     = segment.state.numerics.solver.method,
                                           sense_step = segment.state.numerics.solver.step_size,
                                           iter       = segment.state.numerics.solver.max_evaluations,
                                           tolerance  = segment.state.numerics.solver.tolerance_solution)
    
        # Terminate suppression of console window output   
        sys.stdout = sys.__stdout__  
         
        if outputs[3] != 0:
            mission_converge = False        
            error_message =  outputs[4] 
        else:
            mission_converge = True
     
    elif segment.state.numerics.solver.type  == "root_finder": 
        unknowns = segment.state.unknowns.pack_array() 
         
        unknowns,infodict,ier,error_message = scipy.optimize.fsolve(iterate_root_finder,
                                             unknowns,
                                             args   = segment,
                                             xtol   = segment.state.numerics.solver.tolerance_solution,
                                             maxfev = segment.state.numerics.solver.max_evaluations,
                                             epsfcn = segment.state.numerics.solver.step_size,
                                             full_output = 1)
        
        if ier !=1:
            mission_converge = False
        else:
            mission_converge = True
            
    else: 
        raise Exception('undefined mission solver type')        
        
    if mission_converge == False:
        print("Segment did not converge. Segment Tag: " + segment.tag)
        print("Error Message:\n" + error_message)
        segment.state.numerics.solver.converged = False
        segment.converged = False
    else:
        segment.state.numerics.solver.converged = True
        segment.converged = True
                                
    return
    
# ---------------------------------------------------------------------------------------------------------------------- 
#  Helper Functions
# ---------------------------------------------------------------------------------------------------------------------- 
def iterate_root_finder(unknowns, segment):
    
    """Runs one iteration of of all analyses for the mission.

    Assumptions:
    N/A

    Source:
    N/A

    Inputs:
    state.unknowns                [Data]
    segment.process.iterate       [Data]

    Outputs:
    residuals                     [Unitless]

    Properties Used:
    N/A
    """       
    if isinstance(unknowns,np.ndarray):
        segment.state.unknowns.unpack_array(unknowns)
    else:
        segment.state.unknowns = unknowns
        
    segment.process.iterate(segment)
    
    residuals = segment.state.residuals.pack_array()
        
    return residuals



def add_mission_variables(segment):
    """Make a pretty table view of the problem with objective and constraints at the current inputs for the dummy solver
    

        Assumptions:
        N/A

        Source:
        N/A

        Inputs:
        x                  [vector]

        Outputs:
        input              [array]
        const_table        [array]

        Properties Used:
        None
    """             
    
    # Step 1: Define Nexus
    nexus                        = Nexus()
    optimization_problem         = Data() 
    
    # Step 2 : Get segment type 
    ground_seg_flag =  (type(segment) == RCAIDE.Framework.Mission.Segments.Ground.Landing) or\
                       (type(segment) == RCAIDE.Framework.Mission.Segments.Ground.Takeoff) or \
                       (type(segment) == RCAIDE.Framework.Mission.Segments.Ground.Ground) 
    constant_throttle_seg = type(segment) == RCAIDE.Framework.Mission.Segments.Cruise.Constant_Throttle_Constant_Altitude 
    single_pt_seg = (type(segment) == RCAIDE.Framework.Mission.Segments.Single_Point.Set_Speed_Set_Altitude) or\
                    (type(segment) == RCAIDE.Framework.Mission.Segments.Single_Point.Set_Speed_Set_Altitude_AVL_Trimmed) or \
                    (type(segment) == RCAIDE.Framework.Mission.Segments.Single_Point.Set_Speed_Set_Altitude_No_Propulsion) or \
                    (type(segment) == RCAIDE.Framework.Mission.Segments.Single_Point.Set_Speed_Set_Throttle) 


    
    # Step 2: Optimizer Inputs 
    # Step 2.1: Extract inputs
    input_count  = 0
    unknown_keys = list(segment.state.unknowns.keys())
    unknown_keys.remove('tag') 
    if ground_seg_flag: 
        n_points     = segment.state.numerics.number_of_control_points
        len_inputs = n_points 
    elif constant_throttle_seg:
        n_points     = segment.state.numerics.number_of_control_points  
        len_inputs = n_points*(len(unknown_keys)-1) + 1
    elif single_pt_seg:
        n_points   = 1
        len_inputs = len(unknown_keys)
    else:
        n_points     = segment.state.numerics.number_of_control_points  
        len_inputs = n_points*len(unknown_keys)
        
    unknown_value  = Data()
    full_unkn_vals = Data()
    for unkn in unknown_keys:
        unknown_value[unkn]  = segment.state.unknowns[unkn]
        full_unkn_vals[unkn] = unknown_value[unkn] 

    # Step 2.2: Construct nexus format  : [Variable_###, initial, -np.inf, np.inf , scaling, Units.less]
    initial_values    = full_unkn_vals.pack_array()
    input_len_strings = np.tile('Variable_', len_inputs)
    input_numbers     = np.linspace(1,len_inputs,len_inputs,dtype=np.int16)
    input_names       = np.core.defchararray.add(input_len_strings,np.array(input_numbers+input_count).astype(str))
    bounds            = np.broadcast_to((-np.inf,np.inf),(len_inputs,2))
    units             = np.broadcast_to(Units.less,(len_inputs,))
    new_inputs        = np.reshape(np.tile(np.atleast_2d(np.array([None,None,None,None,None,None])),len_inputs), (-1, 6))
    
    # scaling factor for optimizer 
    factor = np.ceil(np.log10(abs(initial_values)))
    factor[np.isinf(factor)] = 0
    scale  = 10 ** (factor)
    
    # Step 2.4 Add in the inputs 
    new_inputs[:,0]     = input_names   
    new_inputs[:,1]     = initial_values 
    new_inputs[:,2:4]   = bounds   
    new_inputs[:,4]     = scale
    new_inputs[:,5]     = units 
    optimization_problem.inputs = np.array(new_inputs,dtype=object)
    
    # Step 3: Constraints 
    # Step 3.1 : Create the equality constraints to the beginning of the constraints all equality constraints are 0, scale 1, and unitless
    new_con = np.reshape(np.tile(np.atleast_2d(np.array([None,None,None,None,None])),len_inputs), (-1, 5))

    con_len_strings = np.tile('Residual_', len_inputs)
    con_names       = np.core.defchararray.add(con_len_strings,np.array(input_numbers+input_count).astype(str))
    equals          = np.broadcast_to('=',(len_inputs,))
    zeros           = np.zeros(len_inputs)
    ones            = np.ones(len_inputs)
    
    # Step 3.2 Add in the new constraints
    new_con[:,0]    = con_names
    new_con[:,1]    = equals
    new_con[:,2]    = zeros  
    new_con[:,3]    = ones
    new_con[:,4]    = 1*Units.less
    optimization_problem.constraints =  np.array(new_con,dtype=object)            
    
    # Step 4. Aliases 
    # Step 4.1: Setup the aliases for the inputs
    basic_string_con = Data()
    input_string = []

    if ground_seg_flag:       
        output_numbers = np.linspace(0,n_points-2,n_points-1,dtype=np.int16)
        basic_string_con[unknown_keys[1]] = np.tile('segment.state.unknowns.'+unknown_keys[1]+'[', n_points-1)
        input_string.append(np.core.defchararray.add(basic_string_con[unknown_keys[1]],np.array(output_numbers).astype(str)))
        input_string        = np.array(input_string[0])
        input_string        = np.core.defchararray.add(input_string, np.tile(']',len_inputs-1))
        input_aliases       = np.reshape(np.tile(np.atleast_2d(np.array((None,None))),len_inputs), (-1, 2)) 
        input_aliases[:,0]  = input_names
        input_aliases[0,1]  = 'segment.state.unknowns.'+unknown_keys[0] 
        input_aliases[1:,1] = input_string
        
    elif constant_throttle_seg: 
        output_numbers = np.linspace(0,n_points-1,n_points,dtype=np.int16)        
        for unkn in unknown_keys[:-1]:
            basic_string_con[unkn] = np.tile('segment.state.unknowns.'+unkn+'[', n_points)
            input_string.append(np.core.defchararray.add(basic_string_con[unkn],np.array(output_numbers).astype(str)))
        input_string         = np.ravel(input_string)
        input_string         = np.core.defchararray.add(input_string, np.tile(']',len_inputs-1)) 
        input_aliases        = np.reshape(np.tile(np.atleast_2d(np.array((None,None))),len_inputs), (-1, 2)) 
        input_aliases[:,0]   = input_names   
        input_aliases[:-1,1] = input_string 
        input_aliases[-1,1]  = 'segment.state.unknowns.'+unknown_keys[-1]
        
    elif single_pt_seg:  
        for unkn in unknown_keys:
            basic_string_con[unkn] = np.tile('segment.state.unknowns.'+unkn+'[', n_points)
            input_string.append(np.core.defchararray.add(basic_string_con[unkn],np.array([0]).astype(str)))
        input_string       = np.ravel(input_string)
        input_string       = np.core.defchararray.add(input_string, np.tile(']',len_inputs))
        input_aliases      = np.reshape(np.tile(np.atleast_2d(np.array((None,None))),len_inputs), (-1, 2)) 
        input_aliases[:,0] = input_names
        input_aliases[:,1] = input_string    
    else:  
        output_numbers = np.linspace(0,n_points-1,n_points,dtype=np.int16) 
        for unkn in unknown_keys:
            basic_string_con[unkn] = np.tile('segment.state.unknowns.'+unkn+'[', n_points)
            input_string.append(np.core.defchararray.add(basic_string_con[unkn],np.array(output_numbers).astype(str)))
        input_string       = np.ravel(input_string)
        input_string       = np.core.defchararray.add(input_string, np.tile(']',len_inputs))
        input_aliases      = np.reshape(np.tile(np.atleast_2d(np.array((None,None))),len_inputs), (-1, 2)) 
        input_aliases[:,0] = input_names
        input_aliases[:,1] = input_string
    
    # Step 4.2: Setup the aliases for the residuals
    basic_string_res      = np.tile('segment.state.residuals.pack_array()[', len_inputs)
    residual_string       = np.core.defchararray.add(basic_string_res,np.array(input_numbers-1).astype(str))
    residual_string       = np.core.defchararray.add(residual_string, np.tile(']',len_inputs))
    residual_aliases      = np.reshape(np.tile(np.atleast_2d(np.array((None,None))),len_inputs), (-1, 2)) 
    residual_aliases[:,0] = con_names
    residual_aliases[:,1] = residual_string
        
    # Step 4.3: Append Aliases
    aliases = []
    for ii in range(len_inputs):
        aliases.append(input_aliases[ii].tolist()) 
        aliases.append(residual_aliases[ii].tolist())
    
    # Step 5: Objective function
    if segment.state.numerics.solver.objective == None:     
        aliases.append([ 'nothing'                   , 'postprocess.nothing']) 
        optimization_problem.objective = np.array([ [  'nothing'  ,  1   ,    1*Units.less]  ],dtype=object)            
    elif segment.state.numerics.solver.objective == "energy":
        aliases.append([ 'energy_consumed'          , 'postprocess.energy_consumed']) 
        optimization_problem.objective = np.array([ [  'energy_consumed'  ,  1   ,    1*Units.less]  ],dtype=object)            
    elif segment.state.numerics.solver.objective == "power":
        aliases.append([ 'maximum_power'          , 'postprocess.maximum_power'])
        optimization_problem.objective = np.array([ [  'maximum_power'  ,  1   ,    1*Units.less]  ],dtype=object)   
    else:
        raise Exception('undefined objective function')
    
    # append aliases 
    optimization_problem.aliases = aliases        
    
    # Step 6: Expand Rows  
    segment.process.initialize.expand_state(segment)
    
    # Step 7: Update iteration
    input_count = input_count+input_numbers[-1]      
     
    # Step 8: Append segment
    nexus.segment = segment
     
    # Step 9: Append procedure
    nexus.procedure = iterate_segment()
     
    # Step 9: Append post-process
    nexus.postprocess = Data()
    
    # Step 10: Append optimization problem 
    nexus.optimization_problem   = optimization_problem
    
    return nexus

def iterate_segment(): 
    procedure                           = Process()  
    procedure.segment                   = Process()
    procedure.segment.design_mission    = iterate_optimizer     
    procedure.post_process              = segment_post_process   
        
    return procedure
    
def iterate_optimizer(nexus):
    segment = nexus.segment
     
    unknowns = segment.state.unknowns.pack_array()
    if isinstance(unknowns,np.ndarray):
        segment.state.unknowns.unpack_array(unknowns)
    else:
        segment.state.unknowns = unknowns
        
    segment.process.iterate(segment)
    
    residuals = segment.state.residuals.pack_array()    
    nexus.residuals =  residuals
    return nexus


  
def segment_post_process(nexus):
    # unpack
    power      = nexus.segment.state.conditions.energy.power
    I          = nexus.segment.state.numerics.time.integrate
    
    # compute max power of segment 
    max_power  = np.max(nexus.segment.state.conditions.energy.power)
    
    # compute total energy consumed 
    if (type(nexus.segment) == RCAIDE.Framework.Mission.Segments.Single_Point.Set_Speed_Set_Altitude) or\
                    (type(nexus.segment) == RCAIDE.Framework.Mission.Segments.Single_Point.Set_Speed_Set_Altitude_AVL_Trimmed) or \
                    (type(nexus.segment) == RCAIDE.Framework.Mission.Segments.Single_Point.Set_Speed_Set_Altitude_No_Propulsion) or \
                    (type(nexus.segment) == RCAIDE.Framework.Mission.Segments.Single_Point.Set_Speed_Set_Throttle): 
        energy_consumed =  0
    else:
        energy_consumed = np.dot(I,power)[-1][0]
    
    postprocess                 = nexus.postprocess
    postprocess.maximum_power   = max_power
    postprocess.energy_consumed = energy_consumed 
    postprocess.nothing         = 0
    
    return nexus  


# Titanum.py
#
# Created: Jul, 2022, J. Smart
# Modified:

# -------------------------------------------------------------------------------
# Imports
# -------------------------------------------------------------------------------

from .Solid import Solid
from RCAIDE.Framework.Core import Units


# -------------------------------------------------------------------------------
# Aluminum 6061-T6 Solid Class
# ------------------------------------------------------------------------------- 
class Titanium(Solid):
    """
    A class representing Grade 5 Ti-6Al-4V titanium alloy material properties.

    Attributes
    ----------
    ultimate_tensile_strength : float
        Maximum tensile stress before failure in Pa (950e6)
    ultimate_shear_strength : float
        Maximum shear stress before failure in Pa (550e6)
    ultimate_bearing_strength : float
        Maximum bearing stress before failure in Pa (1860e6)
    yield_tensile_strength : float
        Stress at which material begins to deform plastically in Pa (880e6)
    yield_shear_strength : float
        Shear stress at which material begins to deform plastically in Pa (550e6)
    yield_bearing_strength : float
        Bearing stress at which material begins to deform plastically in Pa (1480e6)
    minimum_gage_thickness : float
        Minimum manufacturable thickness in m (0.0)
    density : float
        Material density in kg/m³ (4430)

    Notes
    -----
    This class implements material properties for Ti-6Al-4V (Grade 5) titanium alloy 
    in the annealed condition. This alloy is widely used in aerospace applications 
    due to its excellent strength-to-weight ratio and corrosion resistance.

    **Definitions**
    
    'Ti-6Al-4V'
        A titanium alloy containing 6% aluminum and 4% vanadium, also known as Grade 5
    
    'Ultimate Strength'
        The maximum stress that a material can withstand before failure
    
    'Yield Strength'
        The stress at which a material begins to deform plastically

    References
    ----------
    [1] Aerospace Specification Metals, Inc. (n.d.). Titanium Ti-6Al-4V (Grade 5), Annealed. ASM material data sheet. https://asm.matweb.com/search/SpecificMaterial.asp?bassnum=mtp641 
    """

    def __defaults__(self):
        """Sets material properties at instantiation.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        N/A

        Outputs:
        N/A

        Properties Used:
        None
        """

        self.ultimate_tensile_strength  = 950e6 * Units.Pa
        self.ultimate_shear_strength    = 550e6 * Units.Pa
        self.ultimate_bearing_strength  = 1860e6 * Units.Pa
        self.yield_tensile_strength     = 880e6 * Units.Pa
        self.yield_shear_strength       = 550e6 * Units.Pa
        self.yield_bearing_strength     = 1480e6 * Units.Pa
        self.minimum_gage_thickness     = 0.0 * Units.m
        self.density                    = 4430. * Units['kg/(m**3)']



# RCAIDE/Library/Methods/Powertrain/Propulsors/Internal_Combustion_Engine/design_internal_combustion_engine.py
# 
# Created:  Mar 2025, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE Imports
import RCAIDE 
from RCAIDE.Library.Methods.Powertrain.Converters.Rotor  import design_propeller  
from RCAIDE.Library.Methods.Powertrain                   import setup_operating_conditions 

# Python package imports
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Design Electric Rotor 
# ----------------------------------------------------------------------------------------------------------------------
def design_internal_combustion_engine(ICE, number_of_stations=20, solver_name='SLSQP', iterations=200,
                                     solver_sense_step=1E-5, solver_tolerance=1E-4, print_iterations=False):
    """
    Sizes the propeller of a propeller-driven internal combustion engine and computes
    sea level static performance.
    
    Parameters
    ----------
    ICE : RCAIDE.Library.Components.Propulsors.Internal_Combustion_Engine
        Internal combustion engine propulsor component with the following attributes:
            - tag : str
                Identifier for the propulsor
            - propeller : Data
                Propeller component to be designed
    number_of_stations : int, optional
        Number of radial stations for propeller blade discretization
        Default: 20
    solver_name : str, optional
        Name of the numerical solver to use for propeller design
        Default: 'SLSQP'
    iterations : int, optional
        Maximum number of iterations for the solver
        Default: 200
    solver_sense_step : float, optional
        Step size for finite difference approximations in the solver
        Default: 1E-5
    solver_tolerance : float, optional
        Convergence tolerance for the solver
        Default: 1E-4
    print_iterations : bool, optional
        Flag to print solver iterations
        Default: False
    
    Returns
    -------
    None
        Results are stored in the ICE object:
            - sealevel_static_thrust : float
                Sea level static thrust [N]
            - sealevel_static_power : float
                Sea level static power [W]
    
    Notes
    -----
    This function performs two main tasks:
        1. Designs the propeller using the design_propeller function with the specified
           number of stations
        2. Computes the sea level static performance (thrust and power) of the engine-propeller
           combination at full throttle
    
    The sea level static performance is calculated by:
        * Setting up atmospheric conditions at sea level
        * Creating a low-speed operating state (1% of sea level speed of sound)
        * Setting the throttle to maximum (1.0)
        * Computing the performance at these conditions
    
    **Major Assumptions**
        * US Standard Atmosphere 1976 is used for atmospheric properties
        * Sea level static conditions are approximated with a very low velocity (1% of speed of sound)
        * Full throttle (throttle = 1.0) is used for sea level static performance
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.design_propeller
    RCAIDE.Library.Methods.Powertrain.setup_operating_conditions
    """
    
    # Step 1 Design the Propeller  
    design_propeller(ICE.propeller,number_of_stations = 20) 
     
    # Static Sea Level Thrust   
    atmosphere            = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976() 
    atmo_data_sea_level   = atmosphere.compute_values(0.0,0.0)   
    V                     = atmo_data_sea_level.speed_of_sound[0][0]*0.01 
    operating_state       = setup_operating_conditions(ICE, altitude = 0,velocity_range=np.array([V]))  
    operating_state.conditions.energy.propulsors[ICE.tag].throttle[:,0] = 1.0  
    sls_T,_,sls_P,_,_,_               = ICE.compute_performance(operating_state) 
    ICE.sealevel_static_thrust        = sls_T[0][0]
    ICE.sealevel_static_power         = sls_P[0][0]
    return 


# RCAIDE/Library/Plots/Aerodynamics/plot_rotor_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
import pandas as pd 
import plotly.graph_objects as go
from plotly.subplots import make_subplots 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------      
def plot_rotor_performance(rotor, outputs,title=None, show_figure = True, save_figure=False, save_filename='Rotor_Performance', file_type=".png"):
    """
    Generate plots summarizing rotor aerodynamic performance distributions.

    Parameters
    ----------
    rotor : Data
        Rotor data structure containing outputs with fields:

        - disc_radial_distribution : array
            Radial positions on disc [m]
        - disc_axial_velocity : array
            Total axial velocity [m/s]
        - disc_tangential_velocity : array
            Total tangential velocity [m/s]
        - disc_axial_induced_velocity : array
            Induced axial velocity [m/s]
        - disc_tangential_induced_velocity : array
            Induced tangential velocity [m/s]
        - disc_thrust_distribution : array
            Local thrust distribution [N]
        - disc_torque_distribution : array
            Local torque distribution [N-m]

    title : str, optional
        Custom plot title, default None

    show_figure : bool, optional
        Display figure if True, default True

    save_figure : bool, optional
        Save figure to file if True, default False

    save_filename : str, optional
        Name for saved figure file, default 'Rotor_Performance'

    file_type : str, optional
        File extension for saved figure, default ".png"

    Returns
    -------
    fig : matplotlib.figure.Figure
        Figure containing four subplots:

        - Velocity distributions
        - Induced velocity distributions
        - Thrust distribution
        - Torque distribution

    Notes
    -----
    Each subplot includes:

    - Appropriate axis labels
    - Legend identifying components
    - Consistent line styling

    **Definitions**

    'Total Velocity'
        Sum of freestream and induced velocities
    
    'Induced Velocity'
        Additional velocity induced by rotor
    
    'Thrust Distribution'
        Local thrust force per unit radius
    
    'Torque Distribution'
        Local torque per unit radius

    See Also
    --------
    plot_rotor_disc_performance : Detailed disc visualization
    """
    # unpack 
    r_distribution = outputs.disc_radial_distribution[0, :, 0]
    
    # 2d plots
    fig = make_subplots(rows=2, cols=2)
    
    df1a = pd.DataFrame(dict(x=r_distribution, y=outputs.disc_axial_velocity[0, :, 0])) # label='Axial'
    df1b = pd.DataFrame(dict(x=r_distribution, y=outputs.disc_tangential_velocity[0, :, 0]))  # label='Tangential'
    df2a = pd.DataFrame(dict(x=r_distribution, y=outputs.disc_axial_induced_velocity[0, :, 0])) # label='Axial'
    df2b = pd.DataFrame(dict(x=r_distribution, y=outputs.disc_tangential_induced_velocity[0, :, 0])) # label='Tangential'
    df3  = pd.DataFrame(dict(x=r_distribution, y=outputs.disc_thrust_distribution[0, :, 0]))
    df4  = pd.DataFrame(dict(x=r_distribution, y=outputs.disc_torque_distribution[0, :, 0]))
    
    fig.append_trace(go.Line(df1a, name='Axial', legendgroup='1',showlegend=True), row=1, col=1)
    fig.append_trace(go.Line(df1b, name='Tangential', legendgroup='1',showlegend=True), row=1, col=1)
    fig.append_trace(go.Line(df2a, name='Axial', legendgroup='2',showlegend=True), row=1, col=2)    
    fig.append_trace(go.Line(df2b, name='Tangential', legendgroup='2',showlegend=True), row=1, col=2)     
    fig.append_trace(go.Line(df3, name='Thrust', legendgroup='3',showlegend=False), row=2, col=1)    
    fig.append_trace(go.Line(df4, name='Torque', legendgroup='4',showlegend=False), row=2, col=2)        
    
    fig.update_xaxes(title_text="Radial Station", row=1, col=1)
    fig.update_yaxes(title_text="Velocity", row=1, col=1)
    fig.update_xaxes(title_text="Radial Station", row=1, col=2)
    fig.update_yaxes(title_text="Induced Velocity", row=1, col=2)
    fig.update_xaxes(title_text="Radial Station", row=2, col=1)
    fig.update_yaxes(title_text="Thrust, N", row=2, col=1)
    fig.update_xaxes(title_text="Radial Station", row=2, col=2)
    fig.update_yaxes(title_text="Torque, N-m", row=2, col=2)
    
    fig.update_layout(title_text="Rotor Performance", height=700)
    
    if save_figure:
        fig.write_image(save_filename + '_2D' + file_type) 
    
    if show_figure:
        fig.write_html( save_filename + '.html', auto_open=True)
    return fig 
 


 # RCAIDE/Library/Missions/Segments/Descent/Linear_Mach_Constant_Rate.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------  
#  IMPORT 
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports  
from RCAIDE.Library.Mission.Common.Update.atmosphere import atmosphere

# pacakge imports 
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  
def initialize_conditions(segment):
    """
    Initializes conditions for linear Mach descent at fixed rate

    Parameters
    ----------
    segment : Segment
        The mission segment being analyzed
            - descent_rate : float
                Rate of descent [m/s]
            - mach_number_start : float
                Initial Mach number [-]
            - mach_number_end : float
                Final Mach number [-]
            - altitude_start : float
                Initial altitude [m]
            - altitude_end : float
                Final altitude [m]
            - sideslip_angle : float
                Aircraft sideslip angle [rad]
            - state:
                numerics.dimensionless.control_points : array
                    Discretization points [-]
                conditions : Data
                    State conditions container
                initials : Data, optional
                    Initial conditions from previous segment

    Returns
    -------
    None

    Notes
    -----
    This function sets up the initial conditions for a descent segment with linearly
    varying Mach number and constant descent rate. The true airspeed varies with both
    Mach number and local speed of sound. Updates segment with velocity vector, position
    vector, and altitude.

    **Calculation Process**
        1. Get atmospheric properties for speed of sound
        2. Discretize altitude profile
        3. Calculate Mach number variation:
        M = M0 + (Mf - M0)*t where:
            - M0 is initial Mach number
            - Mf is final Mach number
            - t is normalized time/distance
        4. Calculate true airspeed:
            V = M * a where:
                - M is local Mach number
                - a is local speed of sound
        5. Decompose velocity using:
            - Fixed descent rate
            - Sideslip angle
            - Computed true airspeed

    **Major Assumptions**
        * Linear Mach number variation
        * Constant descent rate
        * Standard atmosphere model
        * Small angle approximations
        * Quasi-steady flight

    See Also
    --------
    RCAIDE.Framework.Mission.Segments
    RCAIDE.Library.Mission.Common.Update.atmosphere
    """      
    
    # unpack
    descent_rate = segment.descent_rate
    M0           = segment.mach_number_start
    Mf           = segment.mach_number_end
    alt0         = segment.altitude_start 
    altf         = segment.altitude_end 
    beta         = segment.sideslip_angle    
    t_nondim     = segment.state.numerics.dimensionless.control_points
    conditions   = segment.state.conditions  

    # Update freestream to get speed of sound
    atmosphere(segment)
    a          = conditions.freestream.speed_of_sound        
    
    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    # check for initial velocity vector
    if M0 is None:
        if not segment.state.initials: raise AttributeError('initial mach number not set')
        M0  =  np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1,:])/a[0,:]         
        
    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0
    conditions.freestream.altitude[:,0] =  alt[:,0]  # positive altitude 

    # process velocity vector
    mach_number = (Mf-M0)*t_nondim + M0
    v_xy_mag    = mach_number * a
    v_z         = descent_rate # z points down
    v_xy        = np.sqrt( v_xy_mag**2 - v_z**2 ) 
    v_x         = np.cos(beta)*v_xy
    v_y         = np.sin(beta)*v_xy
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x[:,0]
    conditions.frames.inertial.velocity_vector[:,1] = v_y[:,0]
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude t



# RCAIDE/Library/Methods/Powertrain/Modulators/__init__.py
# 

"""
Collection of methods for analyzing power modulation components in propulsion systems. 
These methods handle devices that control and modify power delivery, such as Electronic 
Speed Controllers (ESCs). The module provides functionality for computing modulator 
performance and managing power flow between energy sources and mechanical converters.

See Also
--------
RCAIDE.Library.Methods.Powertrain.Propulsors
RCAIDE.Library.Components.Powertrain.Modulators
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .                        import Electronic_Speed_Controller


## @defgroup Library-Plots-Performance-Common Common 
# RCAIDE/Library/Plots/Performance/Common/__init__.py
# 

"""
RCAIDE Common Plotting Utilities

This module provides standardized plotting utilities used across RCAIDE's visualization 
capabilities. It includes functions for consistent axis formatting and plot styling.
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .set_axes            import set_axes
from .plot_style          import plot_style  


# RCAIDE/Methods/Energy/Propulsors/__init__.py
# 

"""
This module provides functionality for modeling complete propulsor systems in powertrains. It includes methods for 
various propulsion technologies including electric rotors, electric ducted fans, internal combustion engines, 
and different turbine-based propulsors.

The propulsors module enables the simulation and analysis of the components that convert mechanical or electrical 
power into thrust, providing the propulsive force for aerospace vehicles.

See Also
--------
RCAIDE.Library.Components.Powertrain.Propulsors
RCAIDE.Library.Methods.Powertrain.Converters
RCAIDE.Library.Methods.Powertrain.Sources
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from . import Constant_Speed_Internal_Combustion_Engine
from . import Electric_Rotor 
from . import Electric_Ducted_Fan 
from . import Internal_Combustion_Engine
from . import Turbofan 
from . import Turbojet 
from . import Turboprop 


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_operating_items_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import  RCAIDE 
from RCAIDE.Framework.Core    import Units ,  Data

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Operating Items Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_operating_items_weight(vehicle):
    """
    Calculate the weight of aircraft operating items using FLOPS methodology for a general aviation aircraft.

    Parameters
    ----------
    vehicle : RCAIDE.Vehicle()
        Vehicle data structure (not directly used in current implementation)

    Returns
    -------
    output : Data()
        Data structure containing:
            - misc : float
                Weight of miscellaneous items [kg], currently set to 0
            - flight_crew : float
                Weight of flight crew and their baggage [kg]
            - flight_attendants : float
                Weight of flight attendants [kg], currently set to 0
            - total : float
                Total operating items weight [kg]

    Notes
    -----
    Calculates weights for crew and operating items for a general aviation aircraft.
    Currently implements a simplified version focused on single-pilot operations.
    
    **Major Assumptions**
        * Single pilot operation (1 flight crew member)
        * Standard flight crew weight of 225 lb including baggage
        * No flight attendants
        * No additional passenger service items
    
    **Theory**

    For general aviation aircraft, the operating items weight is primarily the flight crew weight:

    .. math::
        W_{oi} = N_{fc} * W_{fc}

    Where:
        - W_{oi} is total operating items weight [lb]
        - N_{fc} is number of flight crew (1)
        - W_{fc} is flight crew weight with baggage (225 lb)

    References
    ----------
    [1] NASA. (1979). The Flight Optimization System Weights Estimation Method. 
        NASA Technical Report.
    """
    NFLCR = 1 # Number of flight crew. Assumed to be 1
    WFLCRB = NFLCR * 225  # flight crew and baggage weight

    output                           = Data()
    output.misc = 0
    output.flight_crew               = WFLCRB * Units.lbs
    output.flight_attendants         = 0.0
    output.total                     = output.misc + output.flight_crew + \
                                       output.flight_attendants
    return output





## @ingroup Library-Plots-Energy
# RCAIDE/Library/Plots/Energy/plot_battery_module_conditions.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
## @ingroup Library-Plots-Energy
def plot_battery_module_conditions(results,
                                  save_figure = False,
                                  show_legend = True,
                                  save_filename = "Battery_Module_Conditions_",
                                  file_type = ".png",
                                  width = 11, height = 7):
    """
    Creates a six-panel plot showing various battery module-level conditions throughout flight.

    Parameters
    ----------
    results : Results
        RCAIDE results structure containing segment data and battery module conditions
        
    save_figure : bool, optional
        Flag for saving the figure (default: False)
        
    show_legend : bool, optional
        Flag for displaying plot legend (default: True)
        
    save_filename : str, optional
        Base name of file for saved figure (default: "Battery_Module_Conditions_")
        
    file_type : str, optional
        File extension for saved figure (default: ".png")
        
    width : float, optional
        Figure width in inches (default: 11)
        
    height : float, optional
        Figure height in inches (default: 7)

    Returns
    -------
    fig : matplotlib.figure.Figure

    Notes
    -----
    The function creates a 3x2 subplot containing:
        1. State of Charge (SOC) vs time
        2. Module energy vs time
        3. Module current vs time
        4. Module power vs time
        5. Module voltage vs time
        6. Module temperature vs time
    
    Each segment is plotted with a different color from the inferno colormap.
    Different battery modules are distinguished by different markers.
    
    **Major Assumptions**
    
    * For identical battery modules, only the first module's data is plotted
    * Time is converted from seconds to minutes for plotting
    * Energy is converted to Watt-hours for display
    
    **Definitions**
    
    'SOC'
        State of Charge - the level of charge of the battery module relative to its capacity
    'Module Power'
        Total electrical power of the battery module
    'Module Energy'
        Cumulative energy stored/discharged by the battery module
    """  
    
    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     

    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height)
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))      
    axis_1 = plt.subplot(3,2,1)
    axis_2 = plt.subplot(3,2,2) 
    axis_3 = plt.subplot(3,2,3) 
    axis_4 = plt.subplot(3,2,4)
    axis_5 = plt.subplot(3,2,5) 
    axis_6 = plt.subplot(3,2,6)
     
    for network in results.segments[0].analyses.energy.vehicle.networks: 
        busses  = network.busses 
        for bus in busses:
            for b_i, battery in enumerate(bus.battery_modules):
                if b_i == 0 or bus.identical_battery_modules == False:
                    for i in range(len(results.segments)):  
                        time    = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min    
                        battery_conditions  = results.segments[i].conditions.energy[bus.tag].battery_modules[battery.tag]    
                        module_power          = battery_conditions.power[:,0]
                        module_energy         = battery_conditions.energy[:,0]
                        module_volts          = battery_conditions.voltage_under_load[:,0] 
                        module_current        = battery_conditions.current[:,0]
                        module_SOC            = battery_conditions.cell.state_of_charge[:,0]   
                        module_temperature    = battery_conditions.temperature[:,0]   
                    
                        if b_i == 0 and i ==0:                             
                            axis_1.plot(time, module_SOC, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width, label = battery.tag)
                        else:
                            axis_1.plot(time, module_SOC, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_1.set_ylabel(r'SOC')
                        axis_1.set_ylim([0,1.1])
                        set_axes(axis_1)     
                    
                        axis_2.plot(time, module_energy/Units.Wh, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_2.set_ylabel(r'Energy (W-hr)')
                        set_axes(axis_2) 
                    
                        axis_3.plot(time, module_current, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_3.set_ylabel(r'Current (A)')
                        set_axes(axis_3)  
                    
                        axis_4.plot(time, module_power, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_4.set_ylabel(r'Power (W)')
                        set_axes(axis_4)     
                    
                        axis_5.plot(time, module_volts, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width) 
                        axis_5.set_ylabel(r'Voltage (V)')
                        set_axes(axis_5) 
                    
                        axis_6.plot(time, module_temperature, color = line_colors[i], marker = ps.markers[b_i], linewidth = ps.line_width)
                        axis_6.set_ylabel(r'Temperature, $\degree$C')
                        set_axes(axis_6)  
                  
    if show_legend:      
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4)  
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8) 
    
    # set title of plot 
    title_text   = 'Battery Module Conditions'       
    fig.suptitle(title_text) 
    
    if save_figure:
        plt.savefig(save_filename + battery.tag + file_type)    
    return fig 


# RCAIDE/Methods/Energy/Propulsors/Modulators/compute_esc_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke
 
# ----------------------------------------------------------------------------------------------------------------------
# compute_electric_rotor_performance
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_voltage_out_from_throttle(esc,conditions):
    """
    Computes the output voltage of the Electronic Speed Controller (ESC) based on the 
    throttle setting. Implements a linear voltage modulation model where output voltage 
    is proportionally controlled by the throttle position.

    Parameters
    ----------
    esc : RCAIDE.Library.Components.Energy.Modulators.Electronic_Speed_Controller
        The electronic speed controller component
    esc_conditions : RCAIDE.Framework.Mission.Common.Conditions
        ESC-specific operating conditions
            - throttle : float
                Power modulation setting [0-1]
            - inputs : Conditions
                Input parameters
                    - voltage : float
                        Input voltage [V]
            - outputs : Conditions
                Output parameters
                    - voltage : float
                        Output voltage [V]
    conditions : RCAIDE.Framework.Mission.Common.Conditions
        Flight conditions (not directly used but maintained for API consistency)

    Returns
    -------
    None

    Notes
    -----
    The function performs these operations:
        1. Clamps throttle between 0 and 1
        2. Computes output voltage as a linear function of throttle
        3. Updates ESC conditions with new values

    **Major Assumptions**
        * Linear relationship between throttle and output voltage
        * No voltage drop across the ESC
        * Instantaneous throttle response
        * Perfect voltage regulation
        * Throttle values outside [0,1] are clamped

    **Definitions**
      
    'Voltage Modulation'
        Process of controlling output voltage through throttle position
    """
    esc_conditions = conditions.energy.modulators[esc.tag]
    eta            = esc_conditions.throttle * 1.0
    
    # Negative throttle is bad
    eta[eta<=0.0] = 0.0
    
    # Cap the throttle
    eta[eta>=1.0] = 1.0
    
    # Pack the output
    esc_conditions.outputs.voltage  =eta*esc_conditions.inputs.voltage
    esc_conditions.throttle         = eta 
    
    return
 

# ----------------------------------------------------------------------------------------------------------------------
# compute_current_in_from_throttle
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_current_in_from_throttle(esc,conditions):
    """ The current going into the speed controller
    
        Assumptions:
            The ESC draws current.
        
        Inputs:
            esc_conditions.inputs.currentout [amps]
           
        Outputs:
            outputs.currentin      [amps]
        
        Properties Used:
            esc.efficiency - [0-1] efficiency of the ESC
           
    """
     
    esc_conditions = conditions.energy.modulators[esc.tag]
    eta            = esc_conditions.throttle
    eff            = esc.efficiency
    currentout     = esc_conditions.outputs.current 
    currentin      = currentout*eta/eff # The inclusion of eta satisfies a power balance: p_in = p_out/eff
    
    # Pack 
    esc_conditions.inputs.current   = currentin
    esc_conditions.inputs.power     = esc_conditions.inputs.voltage *currentin
    
    return


# RCAIDE/Library/Components/Propulsors/Converters/PMSM_Generator.py
# 
# 
# Created:  Jan 2025, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------   
# RCAIDE imports  
from .Converter  import Converter
from RCAIDE.Framework.Core                  import Data 
from RCAIDE.Library.Methods.Powertrain.Converters.Generator.append_generator_conditions import  append_generator_conditions

# ----------------------------------------------------------------------------------------------------------------------
#  Generator  
# ----------------------------------------------------------------------------------------------------------------------           
class PMSM_Generator(Converter):
    """
    A electric generator component model for electric propulsion systems.

    Attributes
    ----------
    tag : str
        Identifier for the generator. Default is 'generator'.
        
    resistance : float
        Internal electrical resistance of the generator [Ω]. Default is 0.0.
        
    no_load_current : float
        Current drawn by the generator with no mechanical load [A]. Default is 0.0.
        
    speed_constant : float
        generator speed constant (Kv). Default is 0.0.
        
    rotor_radius : float
        Radius of the generator's rotor [m]. Default is 0.0.
        
    efficiency : float
        Overall generator efficiency. Default is 1.0.
        
    gearbox.gear_ratio : float
        Ratio of output shaft speed to generator speed. Default is 1.0.
        
    power_split_ratio : float
        Ratio of power distribution when generator drives multiple loads. Default is 0.0.
        
    design_torque : float
        Design point torque output [N·m]. Default is 0.0.
        
    interpolated_func : callable
        Function for interpolating generator performance. Default is None.

    Notes
    -----
    The DC_generator class models a direct current electric generator's performance
    characteristics. It accounts for electrical, mechanical, and thermal effects
    including:
    * Internal resistance losses
    * No-load current losses
    * Gearbox losses
    * Speed-torque relationships
    * Power distribution for multiple loads

    **Definitions**

    'Kv'
        generator velocity constant, relating voltage to unloaded generator speed

    'No-load Current'
        Current drawn by generator to overcome internal friction when unloaded
        
    'Power Split Ratio'
        Fraction of total power delivered to primary load in multi-load applications

    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.generator
    """      
    def __defaults__(self):
        """This sets the default values for the component to function.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        None
        """           
        self.tag                 = 'PMSM_generator'
        self.active              = True 
        self.resistance          = 0.0
        self.no_load_current     = 0.0
        self.speed_constant      = 0.0 
        self.efficiency          = 1.0
        self.gearbox             = Data()
        self.gearbox.gear_ratio  = 1.0 
        self.design_torque       = 0.0 
        self.inner_diameter      = 0.0
        self.length_of_path      = 0.0
        self.stack_length        = 0.0
        self.winding_factor      = 0.0
        self.mu_0                = 0.0
        self.mu_r                = 0.0
        self.inverse_calculation = False
        self.interpolated_func   = None
        
    def append_operating_conditions(self,segment,energy_conditions, noise_conditions=None): 
        append_generator_conditions(self,segment,energy_conditions)
        return
    


# RCAIDE/Methods/Energy/Propulsors/Rotor_Design/procedure_setup.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

# RCAIDE Imports 
import RCAIDE 
from RCAIDE.Framework.Core                                                        import Units  
from RCAIDE.Library.Methods.Noise.Frequency_Domain_Buildup.Rotor                  import compute_rotor_noise 
from RCAIDE.Framework.Analyses.Process                                            import Process   
from RCAIDE.Framework.Mission.Common                                              import Conditions
from RCAIDE.Library.Methods.Powertrain.Converters.Rotor.compute_rotor_performance import compute_rotor_performance 

# Python package imports   
import numpy as np 
import scipy as sp  
    
# ----------------------------------------------------------------------------------------------------------------------  
#  Procedure Setup 
# ----------------------------------------------------------------------------------------------------------------------    
def procedure_setup():
    """
    Creates a procedure for rotor blade optimization that defines the sequence of analysis steps.
    
    Parameters
    ----------
    None

    Returns
    -------
    procedure : RCAIDE.Framework.Analyses.Process
        Process object containing the following methods:
            - modify_rotor : function
                Updates blade geometry based on optimization variables
            - hover : function
                Analyzes rotor performance in hover conditions
            - oei : function
                Analyzes rotor performance in one-engine-inoperative conditions
            - cruise : function
                Analyzes rotor performance in cruise conditions (for prop-rotors)
            - post_process : function
                Processes results and computes objective function
    
    Notes
    -----
    This function defines the sequence of analysis steps that are executed during each
    iteration of the rotor blade optimization process. The procedure follows a logical
    flow from geometry modification to performance analysis in different flight conditions,
    and finally to post-processing of results.
    
    The procedure includes the following steps:
        1. modify_blade_geometry: Updates the blade chord and twist distributions based on
           the current optimization variables
        2. run_rotor_hover: Analyzes the rotor performance in hover conditions, computing
           thrust, power, figure of merit, and noise
        3. run_rotor_OEI: Analyzes the rotor performance in one-engine-inoperative conditions
        4. run_rotor_cruise: Analyzes the rotor performance in cruise conditions (for prop-rotors)
        5. post_process: Computes the objective function based on performance metrics and
           checks constraint violations
    
    Each step in the procedure accesses the Nexus object that contains the current state
    of the optimization, including the vehicle configurations, design variables, and
    previously computed results.
    
    **Major Assumptions**
        * The procedure assumes that the blade geometry is parameterized using the
          hyperparameter approach defined in the updated_blade_geometry function
        * Performance analysis includes both aerodynamic and acoustic metrics
        * The objective function balances performance and noise based on user-defined weights
    
    See Also
    --------
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.Design.optimization_setup
    RCAIDE.Library.Methods.Powertrain.Converters.Rotor.compute_rotor_performance
    """
    
    # size the base config
    procedure = Process()
    
    # mofify blade geometry 
    procedure.modify_rotor = modify_blade_geometry
    
    # Run the rotor in hover
    procedure.hover       = run_rotor_hover
    
    # Run the rotor in the oei condition
    procedure.oei         = run_rotor_OEI 

     # Run the rotor in cruise
    procedure.ruise       = run_rotor_cruise

    # post process the results
    procedure.post_process = post_process
        
    return procedure


# ----------------------------------------------------------------------
# Update blade geometry 
# ---------------------------------------------------------------------- 
def modify_blade_geometry(nexus): 
    """ Modifies geometry of prop-rotor blade 
          
          Inputs:  
             nexus     - RCAIDE optmization framework with prop-rotor blade data structure [None]
              
          Outputs:   
             procedure - optimization methodology                                         [None]
              
          Assumptions: 
             N/A 
        
          Source:
             None
    """        
 
    # Pull out the vehicles 
    vehicle_hover     = nexus.vehicle_configurations.hover
    rotor_hover       = vehicle_hover.networks.electric.propulsors.electric_rotor.rotor
    vehicle_oei       = nexus.vehicle_configurations.oei
    rotor_oei         = vehicle_oei.networks.electric.propulsors.electric_rotor.rotor    
    
    if nexus.prop_rotor_flag:  
        vehicle_cruise    = nexus.vehicle_configurations.cruise 
        rotor_cruise      = vehicle_cruise.networks.electric.propulsors.electric_rotor.rotor 
        
    airfoils = rotor_hover.airfoils      
    a_loc    = rotor_hover.airfoil_polar_stations   
        
    # Update geometry of blade
    R       = rotor_hover.tip_radius     
    B       = rotor_hover.number_of_blades 
    r       = rotor_hover.radius_distribution
    c       = updated_blade_geometry(rotor_hover.radius_distribution/rotor_hover.tip_radius ,rotor_hover.chord_r,rotor_hover.chord_p,rotor_hover.chord_q,rotor_hover.chord_t)     
    beta    = updated_blade_geometry(rotor_hover.radius_distribution/rotor_hover.tip_radius ,rotor_hover.twist_r,rotor_hover.twist_p,rotor_hover.twist_q,rotor_hover.twist_t)   
    
    # compute max thickness distribution   
    blade_area    = sp.integrate.cumulative_trapezoid(B*c, r-r[0])
    sigma         = blade_area[-1]/(np.pi*R**2)      
    t_max         = np.zeros(len(c))    
    t_c           = np.zeros(len(c))       
    t_max  = np.zeros(len(c))     
    if len(airfoils.keys())>0:
        for j,airfoil in enumerate(airfoils): 
            a_geo         = airfoil.geometry
            locs          = np.where(np.array(a_loc) == j )
            t_max[locs]   = a_geo.max_thickness*c[locs]   
     
    rotor_hover.chord_distribution          = c
    rotor_hover.twist_distribution          = beta  
    rotor_hover.mid_chord_alignment         = c/4. - c[0]/4.
    rotor_hover.max_thickness_distribution  = t_max 
    rotor_hover.thickness_to_chord          = t_c
    rotor_hover.blade_solidity              = sigma     
    vehicle_hover.store_diff() 
      

    rotor_oei.chord_distribution         = rotor_hover.chord_distribution
    rotor_oei.twist_distribution         = rotor_hover.twist_distribution
    rotor_oei.mid_chord_alignment        = rotor_hover.mid_chord_alignment  
    rotor_oei.max_thickness_distribution = rotor_hover.max_thickness_distribution
    rotor_oei.thickness_to_chord         = rotor_hover.thickness_to_chord 
    rotor_oei.blade_solidity             = rotor_hover.blade_solidity    
    vehicle_oei.store_diff()     
     
    if nexus.prop_rotor_flag: 
        rotor_cruise.chord_distribution         = rotor_hover.chord_distribution
        rotor_cruise.twist_distribution         = rotor_hover.twist_distribution
        rotor_cruise.mid_chord_alignment        = rotor_hover.mid_chord_alignment  
        rotor_cruise.max_thickness_distribution = rotor_hover.max_thickness_distribution
        rotor_cruise.thickness_to_chord         = rotor_hover.thickness_to_chord 
        rotor_cruise.blade_solidity             = rotor_hover.blade_solidity     
        vehicle_cruise.store_diff()  
    
    return nexus    


# ----------------------------------------------------------------------
#   Update blade geometry 
# ---------------------------------------------------------------------- 
def updated_blade_geometry(chi,c_r,p,q,c_t):
    """ Computes planform function of twist and chord distributron using hyperparameters  
          
          Inputs:  
             chi - prop-rotor radius distribution [None]
             c_r - hyperparameter no. 1           [None]
             p   - hyperparameter no. 2           [None]
             q   - hyperparameter no. 3           [None]
             c_t - hyperparameter no. 4           [None] 
                   
          Outputs:       
             x_lin  - function distribution       [None]
              
          Assumptions: 
             N/A 
        
          Source:
              Traub, Lance W., et al. "Effect of taper ratio at low reynolds number."
              Journal of Aircraft 52.3 (2015): 734-747.
              
    """           

    n       = np.linspace(len(chi)-1,0,len(chi))          
    theta_n = n*(np.pi/2)/len(chi)              
    y_n     = chi[-1]*np.cos(theta_n)          
    eta_n   = np.abs(y_n/chi[-1])            
    x_cos   = c_r*(1 - eta_n**p)**q + c_t*eta_n  
    x_lin   = np.interp(chi,eta_n, x_cos)  
    return x_lin 


# ----------------------------------------------------------------------
#   Run the Rotor Hover
# ----------------------------------------------------------------------  
def run_rotor_hover(nexus):
     
    # Unpack    
    network               = nexus.vehicle_configurations.hover.networks.electric
    electric_rotor        = network.propulsors.electric_rotor  
    rotor                 = electric_rotor.rotor
    
    # Setup Test conditions
    alpha                   = rotor.optimization_parameters.multiobjective_aeroacoustic_weight 
    speed                   = rotor.hover.design_freestream_velocity 
    altitude                = np.array([rotor.hover.design_altitude])  
    atmosphere              = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmosphere_conditions   = atmosphere.compute_values(altitude)  
  
    segment                                             = RCAIDE.Framework.Mission.Segments.Segment()  
    conditions                                          = RCAIDE.Framework.Mission.Common.Results()
    conditions.freestream.update(atmosphere_conditions)  
    conditions.frames.inertial.velocity_vector          = np.array([[0.,0.,speed]])   
    conditions.frames.body.transform_to_inertial        = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0., -1.]]]) 
    conditions.frames.wind.transform_to_inertial        = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0.,  1.]]]) 
    conditions.frames.planet.true_course                = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0.,  1.]]])  
    segment.state.conditions                            = conditions
     
    rotor.append_operating_conditions(segment,segment.state.conditions.energy,segment.state.conditions.noise)
    
    rotor_conditions                            = segment.state.conditions.energy.converters[rotor.tag]     
    rotor_conditions.omega                      = (atmosphere_conditions.speed_of_sound*rotor.hover.design_tip_mach)/rotor.tip_radius 
    rotor_conditions.blade_pitch_command[:,0]   = rotor.hover.design_blade_pitch_command
    
    compute_rotor_performance(rotor,conditions)   
     
    # Pack the results
    nexus.results.hover.thrust                       = -segment.state.conditions.energy.converters[rotor.tag].thrust[0,2]  
    nexus.results.hover.torque                       = segment.state.conditions.energy.converters[rotor.tag].torque[0][0]
    nexus.results.hover.power                        = segment.state.conditions.energy.converters[rotor.tag].power[0][0]
    nexus.results.hover.power_c                      = segment.state.conditions.energy.converters[rotor.tag].power_coefficient[0][0]
    nexus.results.hover.thurst_c                     = segment.state.conditions.energy.converters[rotor.tag].thrust_coefficient[0][0]
    nexus.results.hover.omega                        = segment.state.conditions.energy.converters[rotor.tag].omega[0][0]
    nexus.results.hover.max_sectional_cl             = np.max(segment.state.conditions.energy.converters[rotor.tag].lift_coefficient[0]) 
    nexus.results.hover.mean_CL                      = np.mean(segment.state.conditions.energy.converters[rotor.tag].lift_coefficient[0]) 
    nexus.results.hover.figure_of_merit              = segment.state.conditions.energy.converters[rotor.tag].figure_of_merit[0][0]  
    nexus.results.hover.efficiency                   = segment.state.conditions.energy.converters[rotor.tag].efficiency[0][0] 
    nexus.results.hover.conditions                   = conditions  
                
    # microphone locations             
    ctrl_pts                                         = 1 
    theta                                            = rotor.optimization_parameters.noise_evaluation_angle 
    S_hover                                          = np.maximum(altitude[0],20*Units.feet)  
    mic_positions_hover                              = np.array([[0.0 , S_hover*np.sin(theta)  ,S_hover*np.cos(theta)]])      
    
 
    # Run noise model    
    conditions.noise.relative_microphone_locations   = np.repeat(mic_positions_hover[ np.newaxis,:,: ],1,axis=0)
    conditions.aerodynamics.angles.alpha             = np.ones((ctrl_pts,1))* 0. * Units.degrees 
    segment                                          = RCAIDE.Framework.Mission.Segments.Segment() 
    segment.state.conditions                         = conditions
    segment.state.conditions.expand_rows(ctrl_pts)  
    noise                                            = RCAIDE.Framework.Analyses.Noise.Frequency_Domain_Buildup() 
    settings                                         = noise.settings   
    num_mic                                          = len(conditions.noise.relative_microphone_locations[0])  
    conditions.noise.number_of_microphones           = num_mic   
    
    if alpha != 1: 
        compute_rotor_noise(conditions.noise.relative_microphone_locations,rotor,segment,settings)    
        nexus.results.hover.mean_SPL   = np.mean(conditions.noise.converters[rotor.tag].SPL_dBA) 
    else: 
        nexus.results.hover.mean_SPL   = 0  

    return nexus

# ----------------------------------------------------------------------
#   One Engine Inoperative 
# ---------------------------------------------------------------------- 
def run_rotor_OEI(nexus):
    
    # Unpack   
    network               = nexus.vehicle_configurations.oei.networks.electric  
    electric_rotor        = network.propulsors.electric_rotor    
    rotor                 = electric_rotor.rotor
    
    # Setup Test conditions
    speed                 = rotor.oei.design_freestream_velocity 
    altitude              = np.array([rotor.oei.design_altitude]) 
    atmosphere            = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmosphere_conditions = atmosphere.compute_values(altitude)   
  
    segment                                             = RCAIDE.Framework.Mission.Segments.Segment()  
    conditions                                          = RCAIDE.Framework.Mission.Common.Results()
    conditions.freestream.update(atmosphere_conditions)  
    conditions.frames.inertial.velocity_vector          = np.array([[0.,0.,speed]])   
    conditions.frames.body.transform_to_inertial        = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0., -1.]]]) 
    conditions.frames.wind.transform_to_inertial        = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0.,  1.]]]) 
    conditions.frames.planet.true_course                = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0.,  1.]]]) 
    segment.state.conditions                            = conditions
     
    rotor.append_operating_conditions(segment,segment.state.conditions.energy,segment.state.conditions.noise)
                
    rotor_conditions                            =  segment.state.conditions.energy.converters[rotor.tag]     
    rotor_conditions.omega                      = (atmosphere_conditions.speed_of_sound*rotor.oei.design_tip_mach)/rotor.tip_radius 
    rotor_conditions.blade_pitch_command[:,0]   = rotor.oei.design_blade_pitch_command
    
    compute_rotor_performance(rotor,conditions)    
            
    # Pack the results
    nexus.results.oei.thrust       = -segment.state.conditions.energy.converters[rotor.tag].thrust[0,2]  
    nexus.results.oei.torque       = segment.state.conditions.energy.converters[rotor.tag].torque[0][0]
    nexus.results.oei.power        = segment.state.conditions.energy.converters[rotor.tag].power[0][0]
    nexus.results.oei.power_c      = segment.state.conditions.energy.converters[rotor.tag].power_coefficient[0][0]
    nexus.results.oei.omega        = segment.state.conditions.energy.converters[rotor.tag].omega[0][0]
    nexus.results.oei.thurst_c     = segment.state.conditions.energy.converters[rotor.tag].thrust_coefficient[0][0]
    nexus.results.oei.efficiency   = segment.state.conditions.energy.converters[rotor.tag].efficiency[0][0]
    nexus.results.oei.conditions   = conditions 
    
    return nexus

# ----------------------------------------------------------------------
#   Run the Rotor Cruise
# ----------------------------------------------------------------------  
def run_rotor_cruise(nexus):
 
    if nexus.prop_rotor_flag:     
        network         = nexus.vehicle_configurations.cruise.networks.electric  
        electric_rotor  = network.propulsors.electric_rotor
        rotor           = electric_rotor.rotor 
        alpha           = rotor.optimization_parameters.multiobjective_aeroacoustic_weight       
        
        # Setup Test conditions
        speed                 = rotor.cruise.design_freestream_velocity 
        altitude              = np.array([rotor.cruise.design_altitude])  
        atmosphere            = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
        atmosphere_conditions = atmosphere.compute_values(altitude) 
      
        segment                                             = RCAIDE.Framework.Mission.Segments.Segment()  
        conditions                                          = RCAIDE.Framework.Mission.Common.Results()
        conditions.freestream.update(atmosphere_conditions)  
        conditions.frames.inertial.velocity_vector          = np.array([[0.,0.,speed]])   
        conditions.frames.body.transform_to_inertial        = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0., -1.]]]) 
        conditions.frames.wind.transform_to_inertial        = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0.,  1.]]]) 
        conditions.frames.planet.true_course                = np.array([[[1., 0., 0.],[0., 1., 0.],[0., 0.,  1.]]]) 
        segment.state.conditions                            = conditions
          
        rotor.append_operating_conditions(segment,segment.state.conditions.energy,segment.state.conditions.noise)
            
        rotor_conditions                            =  segment.state.conditions.energy.converters[rotor.tag]     
        rotor_conditions.omega                      = (atmosphere_conditions.speed_of_sound*rotor.cruise.design_tip_mach)/rotor.tip_radius 
        rotor_conditions.blade_pitch_command[:,0]   = rotor.cruise.design_blade_pitch_command
        
        compute_rotor_performance(rotor,conditions)   
        
        # Pack the results
        nexus.results.cruise.thrust                      = -segment.state.conditions.energy.converters[rotor.tag].thrust[0,2] 
        nexus.results.cruise.torque                      = segment.state.conditions.energy.converters[rotor.tag].torque[0][0]
        nexus.results.cruise.power                       = segment.state.conditions.energy.converters[rotor.tag].power[0][0]
        nexus.results.cruise.power_c                     = segment.state.conditions.energy.converters[rotor.tag].power_coefficient[0][0]
        nexus.results.cruise.omega                       = segment.state.conditions.energy.converters[rotor.tag].omega[0][0]
        nexus.results.cruise.thurst_c                    = segment.state.conditions.energy.converters[rotor.tag].thrust_coefficient[0][0]
        nexus.results.cruise.max_sectional_cl            = np.max(segment.state.conditions.energy.converters[rotor.tag].lift_coefficient[0]) 
        nexus.results.cruise.mean_CL                     = np.mean(segment.state.conditions.energy.converters[rotor.tag].lift_coefficient[0])  
        nexus.results.cruise.efficiency                  = segment.state.conditions.energy.converters[rotor.tag].efficiency[0][0]
        nexus.results.cruise.conditions                  = conditions  
                    
        # microphone locations            
        ctrl_pts                                         = 1 
        theta                                            = rotor.optimization_parameters.noise_evaluation_angle 
        S_cruise                                         = np.maximum(altitude[0],20*Units.feet)  
        mic_positions_cruise                             = np.array([[0.0 ,S_cruise*np.sin(theta)  ,S_cruise*np.cos(theta)]])      
        
        # Run noise model  
        conditions.noise.relative_microphone_locations   = np.repeat(mic_positions_cruise[ np.newaxis,:,: ],1,axis=0)
        conditions.aerodynamics.angles.alpha             = np.ones((ctrl_pts,1))* 0. * Units.degrees 
        segment                                          = RCAIDE.Framework.Mission.Segments.Segment() 
        segment.state.conditions                         = conditions
        segment.state.conditions.expand_rows(ctrl_pts)  
        noise                                            = RCAIDE.Framework.Analyses.Noise.Frequency_Domain_Buildup() 
        settings                                         = noise.settings   
        num_mic                                          = len(conditions.noise.relative_microphone_locations[0])  
        conditions.noise.number_of_microphones           = num_mic    
        
        if alpha != 1: 
            compute_rotor_noise(conditions.noise.relative_microphone_locations,rotor,segment,settings)  
            nexus.results.cruise.mean_SPL   = np.mean(conditions.noise.converters[rotor.tag].SPL_dBA)   
        else:
            nexus.results.cruise.mean_SPL   = 0  
            
    else:     
        nexus.results.cruise.thrust           = 0.0
        nexus.results.cruise.torque           = 0.0
        nexus.results.cruise.power            = 0.0
        nexus.results.cruise.power_c          = 0.0
        nexus.results.cruise.thurst_c         = 0.0
        nexus.results.cruise.omega            = 0.0
        nexus.results.cruise.max_sectional_cl = 0.0
        nexus.results.cruise.mean_CL          = 0.0
        nexus.results.cruise.efficiency       = 0.0  
        nexus.results.cruise.mean_SPL         = 0.0
        nexus.results.cruise.noise_data       = None   

    return nexus
   

# ----------------------------------------------------------------------
#   Post Process Results to give back to the optimizer
# ----------------------------------------------------------------------   
def post_process(nexus):
     
    rotor                           = nexus.vehicle_configurations.hover.networks.electric.propulsors.electric_rotor.rotor  
    rotor_oei                       = nexus.vehicle_configurations.oei.networks.electric.propulsors.electric_rotor.rotor
    alpha                           = rotor.optimization_parameters.multiobjective_aeroacoustic_weight
    beta                            = rotor.optimization_parameters.multiobjective_performance_weight
    gamma                           = rotor.optimization_parameters.multiobjective_acoustic_weight
    ideal_SPL                       = rotor.optimization_parameters.ideal_SPL_dBA  
    ideal_efficiency                = rotor.optimization_parameters.ideal_efficiency      
    ideal_FoM                       = rotor.optimization_parameters.ideal_figure_of_merit  
    print_iter                      = nexus.print_iterations  
    mean_CL_hover                   = nexus.results.hover.mean_CL
    omega_hover                     = nexus.results.hover.omega
    FM_hover                        = nexus.results.hover.figure_of_merit  
    
    # q to p ratios 
    summary                                 = nexus.summary 
    summary.max_sectional_cl_hover          = nexus.results.hover.max_sectional_cl
    summary.chord_p_to_q_ratio              = rotor.chord_p/rotor.chord_q
    summary.twist_p_to_q_ratio              = rotor.twist_p/rotor.twist_q      
    summary.blade_taper_constraint_1        = rotor.chord_distribution[-1]/rotor.chord_distribution[0]
    summary.blade_taper_constraint_2        = rotor.chord_distribution[-1]/rotor.chord_distribution[0]
    summary.blade_twist_constraint          = rotor.twist_distribution [0] - rotor.twist_distribution [-1] 
    summary.OEI_hover_thrust_power_residual = abs(nexus.results.oei.thrust - rotor.oei.design_thrust) / nexus.results.oei.thrust
            
    # thrust/power residuals  
    if rotor.hover.design_thrust == None:
        summary.hover_thrust_power_residual = abs(nexus.results.hover.power - rotor.hover.design_power) / rotor.hover.design_power
    else: 
        summary.hover_thrust_power_residual = abs(nexus.results.hover.thrust - rotor.hover.design_thrust) /rotor.hover.design_thrust 

    # oei
    if rotor.oei.design_thrust == None:
        summary.oei_thrust_power_residual =  abs(nexus.results.oei.power - rotor.oei.design_power) / rotor.oei.design_power
    else: 
        summary.oei_thrust_power_residual = abs(nexus.results.oei.thrust - rotor.oei.design_thrust) /rotor.oei.design_thrust 
    
        
    if nexus.prop_rotor_flag: 
        if rotor.cruise.design_thrust == None:
            summary.cruise_thrust_power_residual = abs(nexus.results.cruise.power - rotor.cruise.design_power) /rotor.cruise.design_power 
        else: 
            summary.cruise_thrust_power_residual =  abs(nexus.results.cruise.thrust - rotor.cruise.design_thrust) /rotor.cruise.design_thrust    
            
    # -------------------------------------------------------
    # OBJECTIVE FUNCTION
    # -------------------------------------------------------   
    performance_objective  = ((ideal_FoM - FM_hover)/ideal_FoM)*beta +  ((ideal_efficiency - nexus.results.cruise.efficiency)/ideal_efficiency)*(1-beta) 
    
    acoustic_objective     = ((nexus.results.hover.mean_SPL  - ideal_SPL)/ideal_SPL)*gamma  + ((nexus.results.cruise.mean_SPL - ideal_SPL)/ideal_SPL)*(1-gamma) 
 
    summary.objective      = (performance_objective*alpha + acoustic_objective*(1-alpha))  
    

    if nexus.prop_rotor_flag:  
        rotor_cru  = nexus.vehicle_configurations.cruise.networks.electric.propulsors.electric_rotor.rotor         
        summary.max_sectional_cl_cruise = nexus.results.cruise.max_sectional_cl   
        
    # -------------------------------------------------------
    # PRINT ITERATION PERFOMRMANCE
    # -------------------------------------------------------      
    if print_iter:
        print("Aeroacoustic Weight          : " + str(alpha))  
        print("Multiobj. Performance Weight : " + str(beta))  
        print("Multiobj. Acoustic Weight    : " + str(gamma)) 
        print("Performance Obj              : " + str(performance_objective))   
        print("Acoustic Obj                 : " + str(acoustic_objective))  
        print("Aeroacoustic Obj             : " + str(summary.objective))    
        print("Blade Taper                  : " + str(summary.blade_taper_constraint_1))
        print("Hover RPM                    : " + str(omega_hover/Units.rpm))     
        if rotor.hover.design_thrust == None: 
            print("Hover Power                  : " + str(nexus.results.hover.power))  
        if rotor.hover.design_power == None: 
            print("Hover Thrust                 : " + str(nexus.results.hover.thrust))  
        print("Hover Average SPL            : " + str(nexus.results.hover.mean_SPL))    
        print("Hover Tip Mach               : " + str(rotor.hover.design_tip_mach)) 
        print("Hover Thrust/Power Residual  : " + str(summary.hover_thrust_power_residual)) 
        print("Hover Figure of Merit        : " + str(FM_hover))  
        print("Hover Max Sectional Cl       : " + str(summary.max_sectional_cl_hover)) 
        print("Hover Blade CL               : " + str(mean_CL_hover))    
        print("OEI Thrust                   : " + str(nexus.results.oei.thrust)) 
        print("OEI Thrust/Power Residual    : " + str(summary.OEI_hover_thrust_power_residual)) 
        print("OEI Tip Mach                 : " + str(rotor_oei.oei.design_tip_mach))  
        print("OEI Collective (deg)         : " + str(rotor_oei.design_blade_pitch_command/Units.degrees)) 
        if nexus.prop_rotor_flag:    
            print("Cruise RPM                   : " + str(nexus.results.cruise.omega/Units.rpm))    
            print("Cruise Collective (deg)      : " + str(rotor_cru.design_blade_pitch_command/Units.degrees)) 
            if rotor_cru.cruise.design_thrust == None:  
                print("Cruise Power                 : " + str(nexus.results.cruise.power)) 
            if rotor_cru.cruise.design_power == None:  
                print("Cruise Thrust                : " + str(nexus.results.cruise.thrust))   
            print("Cruise Tip Mach              : " + str(rotor_cru.cruise.design_tip_mach))  
            print("Cruise Thrust/Power Residual : " + str(summary.cruise_thrust_power_residual))
            print("Cruise Efficiency            : " + str(nexus.results.cruise.efficiency)) 
            print("Cruise Max Sectional Cl      : " + str(summary.max_sectional_cl_cruise))  
            print("Cruise Blade CL              : " + str(nexus.results.cruise.mean_CL))  
        print("\n\n") 

   
    return nexus    



# RCAIDE/Library/Methods/Weights/Correlation_Buildups/FLOPS/compute_systems_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE
import  RCAIDE 
from RCAIDE.Framework.Core    import Units, Data 

# python imports 
import  numpy as  np
 
# ----------------------------------------------------------------------------------------------------------------------
# Systems Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_systems_weight(vehicle):
    """
    Calculate the systems weight for general aviation aircraft using FLOPS methodology.

    Parameters
    ----------
    vehicle : RCAIDE.Vehicle()
        Vehicle data structure containing:
            - networks : list
                List of propulsion networks
            - wings['main_wing'] : Component
                Main wing data structure containing:
                    - sweeps.quarter_chord : float
                        Quarter chord sweep angle [deg]
                    - areas.reference : float
                        Wing reference area [m^2]
                    - spans.projected : float
                        Wing projected span [m]
                    - flap_ratio : float
                        Flap area to wing area ratio
            - fuselages : list
                List of fuselage components containing:
                    - lengths.total : float
                        Fuselage length [m]
                    - width : float
                        Fuselage width [m]
                    - heights.maximum : float
                        Maximum fuselage height [m]
            - flight_envelope : Component
                Flight envelope data containing:
                    - design_mach_number : float
                        Design cruise Mach number
                    - ultimate_load : float
                        Ultimate load factor
                    - design_range : float
                        Design range [nmi]
            - passengers : int
                Number of passengers
            - design_dynamic_pressure : float
                Design dynamic pressure [Pa]
            - mass_properties.max_takeoff : float
                Maximum takeoff weight [kg]

    Returns
    -------
    output : Data()
        Data structure containing:
            - W_flight_control : float
                Flight control system weight [kg]
            - W_hyd_pnu : float
                Hydraulics and pneumatics weight [kg]
            - W_instruments : float
                Instruments weight [kg]
            - W_avionics : float
                Avionics weight [kg]
            - W_apu : float
                APU weight [kg], currently set to 0
            - W_anti_ice : float
                Anti-ice system weight [kg], currently set to 0
            - W_electrical : float
                Electrical system weight [kg]
            - W_ac : float
                Air conditioning weight [kg], currently set to 0
            - W_furnish : float
                Furnishing weight [kg]
            - W_systems : float
                Total systems weight [kg]

    Notes
    -----
    Calculates weights for all aircraft systems using FLOPS methodology.
    
    **Major Assumptions**
        * Hydraulic system pressure is 3000 psf
        * Single fuselage configuration
        * Pressure ratio for cabin pressure (cruise to sea level) is 0.85. i.e. almostno cabin pressurization
        * Passenger cabin length is 25% of fuselage length
    
    **Theory**

    Flight Controls:
    .. math::
        W_{fc} = 0.404 * S_w^{0.317} * (W_{TO}/1000)^{0.602} * N_{ult}^{0.525} * q^{0.345}

    Instruments:
    .. math::
        W_{in} = 0.48 * A_f^{0.57} * M^{0.5} * (10 + 2.5N_{c} + N_{ew} + 1.5N_{ef})

    Hydraulic and Pneumatics:
    .. math::
        W_{hyd} = 0.57 * (A_f + 0.27 * S_w) * (1 + 0.03N_{ew} + 0.05N_{ef}) * (3000 / P)^{0.35} * M^{0.33}

    Electrical:
    .. math::
        W_{elec} = 92 * L_f^{0.4} * W_f^{0.14} * N_{fuse}^{0.27} * N_{eng}^{0.69} * (1 + 0.044N_{c} + 0.0015N_{pax})

    Avionics:
    .. math::
        W_{av} = 15.8 * D_{range}^{0.1} * N_{fc}^{0.7} * A_f^{0.43}

    Furnishing:
    .. math::
        W_{furn} = 127 * N_{fc} + 44 * N_{pax} + 2.6 * L_f * (W_f + D) * N_{fuse}
    
    Air Conditioning:
    .. math::
        W_{ac} = 3.2 * (A_f * D)^{0.6} + 9 * N_{pax}^{0.83} * M + 0.075 * W_{av}

    Where:
        - W_{fc} is flight controls weight [lb]
        - S_w is wing area [ft^2]
        - W_{TO} is takeoff weight [lb]
        - N_{ult} is ultimate load factor
        - q is dynamic pressure [psf]
        - A_f is fuselage planform area [ft^2]
        - M is design Mach number
        - N_{c} is number of crew
        - N_{ew} is number of wing-mounted engines
        - N_{ef} is number of fuselage-mounted engines
        - P is hydraulic system pressure [psf]
        - L_f is fuselage length [ft]
        - W_f is fuselage width [ft]
        - N_{fuse} is number of fuselages
        - N_{eng} is total number of engines
        - N_{pax} is number of passengers
        - D_{range} is design range [nmi]
        - N_{fc} is number of flight crew
        - D is fuselage depth/height [ft]

    References
    ----------
    [1] NASA. (1979). The Flight Optimization System Weights Estimation Method. 
        NASA Technical Report.
    """ 
    NENG = 0
    FNEW = 0
    FNEF = 0 
    for network in  vehicle.networks:
        for propulsor in network.propulsors:
            if isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Turbofan) or  isinstance(propulsor, RCAIDE.Library.Components.Powertrain.Propulsors.Turbojet):
                NENG += 1 
                if propulsor.wing_mounted: 
                    FNEW += 1  
                else:
                    FNEF += 1
                if 'nacelle' in propulsor:
                    nacelle =  propulsor.nacelle 
                    FNAC    = nacelle.diameter / Units.ft
                else:
                    FNAC    = 0                     
            
    VMAX     = vehicle.flight_envelope.design_mach_number
    SFLAP    = 0
    ref_wing = None 
    for wing in  vehicle.wings:
        if isinstance(wing, RCAIDE.Library.Components.Wings.Main_Wing):
            SFLAP  += wing.areas.reference * wing.flap_ratio / Units.ft ** 2
            ref_wing  =  wing
    
    SW = 0
    if ref_wing == None:
        for wing in  vehicle.wings:
            if SW < wing.areas.reference / Units.ft ** 2:
                ref_wing = wing
                SW = ref_wing.areas.reference / Units.ft ** 2
    DELTA  = 0.85 # Pressure Ratio (cruise to sea level) is assumed to be 0.85
    ULF    = vehicle.flight_envelope.ultimate_load   
    QDIVE  = 1481.35*DELTA*VMAX**2
    DG    = vehicle.mass_properties.max_takeoff / Units.lbs
    WSC   = 0.404*SW**0.317 * (DG/1000)**0.602 * ULF**0.525 * QDIVE**0.345
     
    XL = 0
    WF = 0
    L_fus = 0
    for fuselage in vehicle.fuselages:
        if L_fus < fuselage.lengths.total:
            ref_fuselage = fuselage 
            XL  = fuselage.lengths.total / Units.ft
            WF  = fuselage.width / Units.ft
    FPAREA      = XL * WF
    NPASS       = vehicle.passengers

    
    NFLCR = 1 
    WIN     = 0.48 * FPAREA ** 0.57 * VMAX ** 0.5 * (10 + 2.5 * NFLCR + FNEW + 1.5 * FNEF)  # instrumentation weight

    SW      = vehicle.reference_area / Units.ft ** 2
    HYDR    = 3000  # Hydraulic system pressure
    VARSWP  = 0
    WHYD    = 0.57 * (FPAREA + 0.27 * SW) * (1 + 0.03 * FNEW + 0.05 * FNEF) * (3000 / HYDR) ** 0.35 * \
            (1 + 0.04 * VARSWP) * VMAX ** 0.33  # hydraulic and pneumatic system weight

    NFUSE   = len(vehicle.fuselages)
    WELEC   = 92. * XL ** 0.4 * WF ** 0.14 * NFUSE ** 0.27 * NENG ** 0.69 * \
            (1. + 0.044 * NFLCR + 0.0015 * NPASS)  # electrical system weight

    DESRNG  = vehicle.flight_envelope.design_range / Units.nmi
    WAVONC  = 15.8 * DESRNG ** 0.1 * NFLCR ** 0.7 * FPAREA ** 0.43  # avionics weight

    XLP     = 0.25 * XL # Assumption that pax cabin is 25% of fuselage length
    DF      = ref_fuselage.heights.maximum / Units.ft # D stands for depth
    WFURN   = 127 * NFLCR +  44 * vehicle.passengers \
                + 2.6 * XLP * (WF + DF) * NFUSE  # furnishing weight

    WAC     = (3.2 * (FPAREA * DF) ** 0.6 + 9 * NPASS ** 0.83) * VMAX + 0.075 * WAVONC  # ac weight



    output                      = Data()
    output.W_flight_control    = WSC * Units.lbs
    output.W_hyd_pnu           = WHYD * Units.lbs
    output.W_instruments       = WIN * Units.lbs
    output.W_avionics          = WAVONC * Units.lbs
    output.W_apu               = 0.0
    output.W_anti_ice          = 0.0
    output.W_electrical        = WELEC * Units.lbs
    output.W_ac                = 0.0
    output.W_furnish           = WFURN * Units.lbs
    output.W_systems           = WSC  + WIN + WHYD + WELEC + WAVONC + WFURN + WAC 
    return output